/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.document;

import gvid.*;
import java.util.LinkedList;
import java.util.Iterator;

public class Line 
	implements DirtManager,
			   DirtListener {
    private String m_line = null;
    private String m_original = null;
    private int m_firstWord = -1;
    private int m_column = 0;
    
    public Line(String line) {
    	m_dirtImpl.addDirtListener(this);
        m_line = line;
        m_original = null;
        updateDirty();
    }
    
    private Line( String line, int column ) {
    	m_dirtImpl.addDirtListener(this);
        m_line = line;
        m_original = null;
        m_column = column;
        updateDirty();
    }
    
    public Line split( int column ) {
    	Line rc = new Line(m_line.substring(column));
    	rc.m_dirtImpl.copyListeners(m_dirtImpl);
    	rc.removeDirtListener(this);
    	m_line = m_line.substring(0,column);
        updateDirty();
    	return rc;
    }
    
    public void append( Line line ) {
    	m_line += line.m_line;
        updateDirty();
    }
    
    public void insertChar( char c, int column ) {
    	column = Math.min(column, length());
    	m_line = m_line.substring(0,column)+c+m_line.substring(column);
        updateDirty();
    }
    
    public void deleteChar( int column ) {
    	if( column < m_line.length()-1 ) {
    		m_line = m_line.substring(0, column) + m_line.substring(column+1);
    	} else {
    		m_line = m_line.substring(0, column);
    	}
        updateDirty();
    }
    
    public void deleteRange( int start, int end ) {
    	int delEnd = Math.min(Math.max(start,end), m_line.length());
    	int delStart = Math.max(Math.min(start,end),0);
    	m_line = m_line.substring(0, delStart)+m_line.substring(delEnd);
        updateDirty();
    }
    
    public String getText( int start, int end ) {
    	// check bounds
    	int txtEnd = Math.min( end, length() );
    	int txtStart = Math.min(start, end);
    	return m_line.substring(txtStart, txtEnd);
    }
    
    public boolean isEOL(int col) {
        boolean rc = false;
        if( col >= length() ) {
            rc = true;
        }
        return rc;
    }
    
    public int firstWord() {
        if( m_firstWord == -1 ) {
            for( m_firstWord = 0; m_firstWord<m_line.length(); m_firstWord++ ) {
                if( !Character.isWhitespace(m_line.charAt(m_firstWord)) ) {
                    break;
                }
            }
        }
        return m_firstWord;
    }
    
    public int nextWord(int fromIndex) {
        int rc = fromIndex;
        if( rc >= 0 && rc < m_line.length() ) {
            // Get the character type on which we are starting
            char initChar = initChar = m_line.charAt(rc++);
            
            // Move forward until we are no longer on this type or character
            while( rc < m_line.length() ) {
                char c = m_line.charAt(rc);
                if( ( Character.isLetterOrDigit(c) && !Character.isLetterOrDigit(initChar) ) ||
                    ( Character.isWhitespace(c) && !Character.isWhitespace(initChar) ) ||
                    ( !Character.isLetterOrDigit(c) && !Character.isWhitespace(c) && ( Character.isLetterOrDigit(initChar) || Character.isWhitespace(initChar) ) ) ) {
                    // we found a different character
                    break;
                }
                rc++;
            }
            
            // Move forward until we find a non-whitespace character
            while( rc < m_line.length() ) {
                char c = m_line.charAt(rc);
                if( !Character.isWhitespace(c) ) {
                    break;
                }
                rc++;
            }
        } else {
            rc = m_line.length();
        }
        // return the current value
        return rc;
    }
    
    public int prevWord(int fromIndex) {
        int rc = fromIndex;
        if( rc > 0 && rc <= m_line.length() ) {
            // Get the character type on which we are starting
            char initChar = ' ';
            if( rc > 0 && rc < m_line.length() ) {
                initChar = m_line.charAt(--rc);
            } else {
//                rc--;
            }
            
            // Move back until we are no longer on this type or character
            while( rc >= 0 ) {
            	rc--;
            	if( rc >= 0 ) { 
                	char c = m_line.charAt(rc);
                	if( ( Character.isLetterOrDigit(c) && !Character.isLetterOrDigit(initChar) ) ||
                    	( Character.isWhitespace(c) && !Character.isWhitespace(initChar) ) ||
                    	( !Character.isLetterOrDigit(c) && !Character.isWhitespace(c) && ( Character.isLetterOrDigit(initChar) || Character.isWhitespace(initChar) ) ) ) {
                    	// we found a different character
                    	break;
                	}
            	}
            }
            
            if( Character.isWhitespace(initChar) ) {
            	// Move forward until we find a non-whitespace character
            	char wordEnd = ' ';
            	while( rc >= 0 ) {
                	char c = m_line.charAt(rc);
                	if( !Character.isWhitespace(c) ) {
                    	wordEnd = c;
                    	break;
                	}
                	rc--;
            	}
	            
            	// Move to the beginning of the word
            	while( rc >= 0 ) {
                	char c = m_line.charAt(rc);
                	if( Character.isWhitespace(c) ) {
                    	break;
                	} else if( Character.isLetterOrDigit(c) && !Character.isLetterOrDigit(wordEnd) ) {
                    	// the word is not letters and digits and we found a character that is
                    	break;
                	} else if( !Character.isLetterOrDigit(c) && Character.isLetterOrDigit(wordEnd) ) {
                    	// the word is letters and digits and we found a character that isn't
                    	break;
                	}
                	rc--;
            	}
            }

            // return the current value plus one
        } else {
            rc = -1;
        }
        return rc+1;
    }
    
    public int wordStart( int column ) {
    	int rc = column;
        rc = Math.max(rc, 0);
        rc = Math.min(rc, m_line.length());
        while( rc > 0 ) {
        	char front = rc>0 ? m_line.charAt(rc-1) : ' ';
//        	char back = rc<m_line.length() ? m_line.charAt(rc) : ' ';
        	
        	if( Character.isLetterOrDigit(front) || front=='_' ) {
        		rc--;
        	} else {  
        		break;
        	}        	        	        	
          }
    	return rc;
    }
    
    public int wordEnd( int column ) {
    	int rc = column;
        rc = Math.max(rc, 0);
        rc = Math.min(rc, m_line.length());
        while( rc < m_line.length() ) {
//        	char front = rc>0 ? m_line.charAt(rc-1) : ' ';
        	char back = rc<m_line.length() ? m_line.charAt(rc) : ' ';
        	
        	if( Character.isLetterOrDigit(back)  || back=='_' ) {
        		rc++;
        	} else {  
        		break;
        	}        	        	        	
        }
    	return rc;
    }
    
    public int length() {
        return m_line.length();
    }
    
    public String toString() {
        return m_line;
    }
    
    public int indexOf(char c, int fromIndex) {
        return m_line.indexOf(c, fromIndex);
    }
    
    public Line substring( int beginIndex ) {
    	Line rc = new Line( m_line.substring(beginIndex), m_column+beginIndex);
    	rc.m_dirtImpl.copyListeners(m_dirtImpl);
    	rc.removeDirtListener(this);
        return rc;
    }
    
    public Line substring( int beginIndex, int endIndex ) {
        Line rc = new Line( m_line.substring(beginIndex, endIndex), m_column+beginIndex);
    	rc.m_dirtImpl.copyListeners(m_dirtImpl);
    	rc.removeDirtListener(this);
        return rc;
    }
    
    public int charAt(int index) {
    	int rc = -1;
    	if( index>=0 && index < m_line.length() ) {
        	rc = m_line.charAt(index);
    	}
    	return rc;
    }

///////////////////////////////////////////////////////////////////////////////
// DirtManager
	DirtManagerImpl m_dirtImpl = new DirtManagerImpl(this);
    
	public boolean isDirty() { 
		return m_dirtImpl.isDirty();
	}

	public void clearDirty() {
		m_dirtImpl.clearDirty();
	}
	
	public void addDirtListener( DirtListener l ) {
		m_dirtImpl.addDirtListener(l);
	}
	
	public void removeDirtListener( DirtListener l ) {
		m_dirtImpl.removeDirtListener(l);
	}
    
    protected void updateDirty() {
    	if( m_line.equals(m_original) ) {
    		m_dirtImpl.clearDirty();
    	} else {
    		m_dirtImpl.setDirty();
    	}
    }    
// DirtManager
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// DirtListener
	public void dirtyChanged( boolean bDirty, DirtManager mgr ) {
		m_firstWord = -1;
	}
// DirtListener
///////////////////////////////////////////////////////////////////////////////
}