/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.document;

import gvid.*;
import java.awt.Point;
import java.util.*;
import gvid.ui.splitscrollpane.*;
import java.io.*;
import gvid.util.*;
import gvid.undo.*;
import gvid.ui.editor.undo.*;

public class SourceCodeModel 
    implements	SplitScrollModel,
    			DirtListener,
    			DirtManager
{
    private Vector m_lineBuffer = new Vector();
    private Vector m_viewlist = new Vector();
    private ReferencePointManager m_refPoints = new ReferencePointManager();
	private UndoRedoManager m_undoManager = new UndoRedoManager();
	private File m_file = null;
    
    public SourceCodeModel() {
    	addLine(new Line(""));
    	clearDirty();
    }
    
    public SourceCodeModel( File file ) {
    	m_file = file;
    	loadFile();
    }
    
    public File getFile() {
    	return m_file;
    }
    
    public void setFile( File file ) {
    	m_file = file;
    }
    
    public UndoRedoManager getUndoManager() {
    	return m_undoManager;
    }
    
    public boolean addReferencePoint( Point p, String name, ReferencePointListener l ) {
    	return m_refPoints.addReference(p, name, l);
    }
    
    public void removeReferencePoint( String name, ReferencePointListener l ) {
    	m_refPoints.removeReference(name, l);
    }
    
    public String getUnusedReferenceId() {
    	return m_refPoints.getUnusedString();
    }
    
    public void moveReferencePointFrom(String name, SourceCodeModel model) {    	
    	if( model != null ) {
    		m_refPoints.moveReferencePointFrom(name, model.m_refPoints);
    	}
    }
    
    public Point getReferencePoint( String name ) {
    	return m_refPoints.getPoint(name);
    }
    
    public void setReferencePoint( Point p, String name, Object caller ) {
    	m_refPoints.setPoint(p, name, caller);
    }
        
    public int charAt(Point doc) {
    	int rc = -1;
    	Line line = line(doc.y);
    	if( line != null ) {
    		rc = (char)line.charAt(doc.x);
    	}
    	return rc;
    }
    
    private Line line(int index) {
        Line rc = null;
        if( index>=0 && index <m_lineBuffer.size() ) {
            rc = (Line)m_lineBuffer.elementAt(index);;
        }
        return rc;
    }
    
    public int lineFirstWord( int row ) {
    	int rc = 0;
    	Line line = line(row);
    	if( line != null ) {
    		rc = line.firstWord();
    	}
    	return rc;
    }
    
    public String lineToString( int row ) {
    	String rc = "";
    	Line line = line(row);
    	if( line != null ) {
    		rc = line.toString();
    	}
    	return rc;
    }
    
    public int lineIndexOf( int row, char c, int fromIndex ) {
    	int rc = -1;
    	Line line = line(row);
    	if( line != null ) {
    		rc = line.indexOf(c, fromIndex);
    	}
    	return rc;
    }
    
    public int lineLength(int row) {
    	int rc = 0;
    	Line line = line(row);
    	if( line != null ) {
    		rc = line.length();
    	}
    	return rc;
    }
        
    public int lineWordStart( int row, int fromIndex ) {
    	int rc = 0;
    	Line line = line(row);
    	if( line != null ) {
    		rc = line.wordStart(fromIndex);
    	}
    	return rc;
    }
    
    public int lineWordEnd( int row, int fromIndex ) {
    	int rc = 0;
    	Line line = line(row);
    	if( line != null ) {
    		rc = line.wordEnd(fromIndex);
    	}
    	return rc;
    }
    
    public UndoCommand addLine( Line line ) {
    	UndoCommand rc = new InsertLineCommand( "addLine", this, m_lineBuffer.size()-1, line.toString() );
        m_lineBuffer.addElement(line);
        line.addDirtListener(this);
		if( line.isDirty() ) {
    		m_nDirtyLineCount++;
			if( m_nDirtyLineCount == 1 ) {
				notifyDirty();
			}
		}
        return rc;
    }
    
    public UndoCommand split( Point p ) {
    	UndoCommand rc = null;
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
    		Line currentLine = line(p.y);
    		if( currentLine != null ) {
	    		Line newLine = currentLine.split(p.x);
	    		m_lineBuffer.insertElementAt(newLine, p.y+1);
	    		newLine.addDirtListener(this);
				if( newLine.isDirty() ) {
    				m_nDirtyLineCount++;
					if( m_nDirtyLineCount == 1 ) {
						notifyDirty();
					}
				}
	    		m_refPoints.splitRowAt(p);
	    		repaintViews();
	    		rc = new SplitLineCommand("splitLine", this, p);
    		}
    		m_refPoints.removeLock();
    	}    	
    	return rc;
    }
    
    
    public UndoCommand insertChar( char c, Point p ) {
    	InsertCharCommand rc = null;
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
    		Line currentLine = line(p.y);
    		if( currentLine != null ) {
	    		currentLine.insertChar(c, p.x);
	    		m_refPoints.incrementColumnsAfter(p,1);
	    		repaintViews();
    		}
    		rc = new InsertCharCommand("insertChar", this, c, p);
    		m_refPoints.removeLock();
    	}    	
    	return rc;
    }
    
    public UndoCommand deleteChar( Point p ) {
    	UndoCommand rc = null;
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
    		Line currentLine = line(p.y);
    		if( currentLine != null ) {
    			if( currentLine.isEOL(p.x) ) {
    				joinLine(p.y);
    				rc = new JoinLineCommand( "deleteChar", this, p );
                } else {
                	char c = (char)currentLine.charAt(p.x);
	    			currentLine.deleteChar(p.x);
	    			m_refPoints.decrementColumnsAfter(p,1);
	    			repaintViews();
	    			rc = new DeleteCharCommand( "deleteChar", this, c, p );
                }
    		}
    		m_refPoints.removeLock();
    	}    	
    	return rc;
    }
    
    public UndoCommand overwriteChar( char c, Point p ) {
    	UndoChain rc = new UndoChain("overwriteChar");
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
    		Line currentLine = line(p.y);
    		if( currentLine != null ) {
    			if( p.x < currentLine.length() ) {
    				char delChar = (char)currentLine.charAt(p.x);
    				currentLine.deleteChar(p.x);    		
    				m_refPoints.decrementColumnsAfter(p, 1);
    				rc.addRedo(new DeleteCharCommand("overwriteChar", this, delChar, p));
    			}
	    		currentLine.insertChar(c, p.x);
	    		m_refPoints.incrementColumnsAfter(p,1);
	    		repaintViews();
    		}
    		rc.addRedo(new InsertCharCommand("overwriteChar", this, c, p));
    		m_refPoints.removeLock();
    	}    	
    	return rc;
    }
    
    // Joins the line at index and the next line
    public UndoCommand joinLine( int row ) {
    	UndoCommand rc = null;
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
    		Line currentLine = line(row);
    		Line nextLine = line(row+1);
    		if( currentLine != null && nextLine != null ) {
    			rc = new JoinLineCommand("joinLine", this, new Point(currentLine.length(), row));
    			m_refPoints.joinRowAt(row, currentLine.length());
    			m_lineBuffer.remove(row+1);
    			nextLine.removeDirtListener(this);
    			if( nextLine.isDirty() ) {
    				m_nDirtyLineCount--;
					if( m_nDirtyLineCount == 0 ) {
						notifyDirty();
					}
    			}
    			currentLine.append(nextLine);
    			repaintViews();
    		}
    		m_refPoints.removeLock();
    	}  
    	return rc;
    }
    
    public UndoCommand removeLine( int row ) {
    	UndoCommand rc = null;
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
	    	if( ( row >= 0 ) && ( row < m_lineBuffer.size() ) ) {
	    		rc = new DeleteLineCommand("removeLine", this, row, line(row).toString() );
	    		m_refPoints.decrementColumnsAfter(new Point(0, row), line(row).length());
   				m_refPoints.joinRowAt(row, line(row).length());
   				repaintViews();
   				Line l = line(row);
   				l.removeDirtListener(this);
				if( l.isDirty() ) {
    				m_nDirtyLineCount--;
					if( m_nDirtyLineCount == 0 ) {
						notifyDirty();
					}
				}
	    		m_lineBuffer.remove(row);
	    	}
    		m_refPoints.removeLock();
    	}  
    	return rc;
    }
    
    public UndoCommand insertLine( int row, String str ) {
    	UndoCommand rc = null;
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
	    	if( ( row >= 0 ) && ( row <= m_lineBuffer.size() ) ) {
	    		rc = new InsertLineCommand("insertLine", this, row, str );
	    		m_refPoints.addLock();
	    		Line line = new Line(str);
	    		line.addDirtListener(this);
    			m_nDirtyLineCount++;
				if( m_nDirtyLineCount == 1 ) {
					notifyDirty();
				}
        		m_lineBuffer.insertElementAt(line, row);
        		m_refPoints.incrementRowsAfter(row, 1);
        		repaintViews();
    			m_refPoints.removeLock();
	    	}
    		m_refPoints.removeLock();
    	}  
    	return rc;
    }
    
    public UndoCommand deleteRange( Point start, Point end ) {
    	UndoChain rc = new UndoChain("deleteRange");
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
    		if( start.y == end.y ) {
    			Line line = line(start.y);
    			if( line != null ) {
    				// update the reference position
    				int front = Math.min(start.x, end.x);
    				int back = Math.max(start.x, end.x);
   					m_refPoints.decrementColumnsAfter(new Point(front, start.y), Math.min(Math.abs(start.x-end.x), line.length()-front));
    				// update the document
    				rc.addRedo( new DeleteLineRangeCommand( "deleteRange", this, start.y, start.x, end.x, line.getText(front, back) ) );
    				line.deleteRange(front, back);
    			}
    		} else {
    			// make sure the points are ordered from beginning to end
    			if( end.y < start.y ) {
    				Point temp = end;
    				end = start;
    				start = temp;
    			}
	    		
    			// delete in the first line
    			Line startLine = line(start.y);
    			if( startLine != null ) {
    				rc.addRedo( new DeleteLineRangeCommand( "deleteRange", this, start.y, start.x, startLine.length(), startLine.getText(start.x, startLine.length()) ) );
    				m_refPoints.decrementColumnsAfter(start, startLine.length()-start.x);
    				startLine.deleteRange(start.x, startLine.length());
    			}
	    		
    			// delete in the last line
    			Line endLine = line(end.y);
    			if( endLine != null ) {
    				rc.addRedo( new DeleteLineRangeCommand( "deleteRange", this, end.y, 0, end.x, endLine.getText(0, end.x) ) );
    				m_refPoints.decrementColumnsAfter(new Point(0,end.y), end.x);
    				endLine.deleteRange(0, end.x);
    			}
	    		
    			// delete the middle lines
    			m_refPoints.decrementRowsAfter(start.y, end.y-start.y-1);
    			for( int i=start.y+1; i<end.y; i++) {
    				rc.addRedo( new DeleteLineCommand( "deleteRange", this, start.y+1, line(start.y+1).toString() ) );
    				Line line = line(start.y+1);
    				line.removeDirtListener(this);
					if( line.isDirty() ) {
    					m_nDirtyLineCount--;
						if( m_nDirtyLineCount == 0 ) {
							notifyDirty();
						}
					}
    				m_lineBuffer.remove(start.y+1);
    			}
	    		
    			rc.addRedo(joinLine(start.y));
    		}
    		m_refPoints.removeLock();
    	}    	
    	return rc;
    }
    
    public UndoCommand insertText( String str, Point start ) {
    	UndoChain rc = new UndoChain("insertText");
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
    		Point currentPos = new Point(start);
    		Line currentLine = line(currentPos.y);
    		for( int i=0; i<str.length(); i++ ) {
    			char c = str.charAt(i);
    			if( c == '\n' ) {
    				// new line - insert the new line
    				rc.addRedo(split(currentPos));
    				currentPos.x = 0;
    				currentPos.y++;
    				currentLine = line(currentPos.y);
    			} else if( c == '\r' ) {
    				// ignore - '\n' will handle line feeds only
    			} else {
    				// insert the character into the current line
    				rc.addRedo(insertChar(c, currentPos));
    				currentPos.x++;
    			}
    		}
    		repaintViews();
    		m_refPoints.removeLock();
    	}
    	return rc;
    }
    
    public void insertLine( String str, int row ) {
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
	    	Line line = new Line(str);
	    	line.removeDirtListener(this);
    		m_nDirtyLineCount++;
			if( m_nDirtyLineCount == 1 ) {
				notifyDirty();
			}
        	m_lineBuffer.insertElementAt(line, row);
        	m_refPoints.incrementRowsAfter(row, 1);
        	repaintViews();
    		m_refPoints.removeLock();
    	}
    }
    
    /**
     * Moves the text range starting at start and ending at end to the new destination. dest
     * is updated to the new document position after the move. The point returned is the new
     * end of the selection.
     * 
     * @param start
     * @param end
     * @param dest
     */
    public Point moveRange( Point start, Point end, Point dest ) {
    	Point rc = null;
    	synchronized( m_lineBuffer ) {
	    	m_refPoints.addLock();
	    	
    		String str = getText( start, end );
    		String uniqueStr = m_refPoints.getUnusedString();
	    	
    		m_refPoints.addReference( dest, uniqueStr, null );
    		deleteRange( start, end );
    		dest.setLocation( m_refPoints.getPoint(uniqueStr) );	    	
	    	m_refPoints.removeReference( uniqueStr, null);
	    	
    		rc = new Point( dest );
    		m_refPoints.addReference( rc, uniqueStr, null );
    		insertText( str, dest );
    		rc.setLocation( m_refPoints.getPoint(uniqueStr) );	    	
	    	m_refPoints.removeReference( uniqueStr, null);
    		repaintViews();
    		
    		m_refPoints.removeLock();
    	}
    	return rc;
    }
    
    public String getText( Point start, Point end ) {
    	StringBuffer rc = new StringBuffer();
    	if( start.y == end.y ) {
    		Line line = line(start.y);
    		if( line != null ) {
    			rc.append(line.getText(Math.min(start.x, end.x), Math.max(start.x, end.x)));
    		}
    	} else {
    		// make sure the points are ordered from beginning to end
    		if( end.y < start.y ) {
    			Point temp = end;
    			end = start;
    			start = temp;
    		}
    		
    		// append from the first line
    		Line line = line(start.y);
    		if( line != null ) {
    			rc.append(line.getText(start.x, line.length())+"\n");
    		}
    		
    		// append the middle lines
    		for( int i=start.y+1; i<end.y; i++) {
    			line = line(i);
    			if( line != null ) {
    				rc.append(line.toString()+"\n");
    			}
    		}
    		
    		// append from the last line
    		line = line(end.y);
    		if( line != null ) {
    			rc.append(line.getText(0, end.x));
    		}
    	}
    	return rc.toString();
    }
    
    public boolean isEOL(Point p ) {
        boolean rc = true;
        if( p.y>=getRowMinimum() && p.y<=getRowMaximum() ) {
            rc = ((Line)m_lineBuffer.elementAt(p.y)).isEOL(p.x);
        } else if( p.y<getRowMinimum() ) {
            rc = false;
        }
        return rc;
    }
    
    public Point nextWord(Point p ) {
        Point rc = new Point(p);
        if( rc.x >= line(rc.y).length() ) {
            if( rc.y < getRowMaximum() ) {
                rc.y++;
                rc.x = line(rc.y).firstWord();
            }
        } else {
            rc.x = line(rc.y).nextWord(rc.x);
        }
        return rc;
    }
    
    public Point prevWord(Point p ) {
        Point rc = new Point(p);
        if( rc.x <= 0 ) {
            if( rc.y > getRowMinimum() ) {
                rc.y--;
                rc.x = line(rc.y).prevWord(line(rc.y).length());
            }
        } else {
            rc.x = line(rc.y).prevWord(rc.x);
        }
        return rc;
    }
    
    public Point wordStart( Point p ) {
        Point rc = new Point(p);
        if( rc.x <= 0 ) {
            if( rc.y < getRowMinimum() ) {
                rc.y--;
                rc.x = line(rc.y).wordStart(line(rc.y).length());
            }
        } else {
            rc.x = line(rc.y).wordStart(rc.x);
        }
        return rc;
    }
    
    public Point wordEnd( Point p ) {
        Point rc = new Point(p);
        if( rc.x >= line(rc.y).length() ) {
            if( rc.y < getRowMaximum() ) {
                rc.y++;
                rc.x = line(rc.y).wordEnd(0);
            }
        } else {
            rc.x = line(rc.y).wordEnd(rc.x);
        }
        return rc;
    }
    
    public void readFile( BufferedReader reader )  
    	throws java.io.IOException {
		String currentLine = reader.readLine();
		while( currentLine != null ) {
			addLine(new Line(currentLine));
			currentLine = reader.readLine();
		}
		clearDirty();
    }
    
    public void writeFile( Writer writer ) 
    	throws java.io.IOException {
    	for( int i=0; i < m_lineBuffer.size(); i++ ) {
    		Line line = (Line)m_lineBuffer.elementAt( i );
    		writer.write( line.toString() );
    		if( i == m_lineBuffer.size()-1 ) {
    			writer.write( "\n" );
    		}
    	}
    	clearDirty();
    }
    
	public boolean loadFile() {
		boolean rc = false;
		if( m_file != null ) {
			try {
				FileReader fr = new FileReader(m_file);
				BufferedReader br = new BufferedReader(fr);
				readFile(br);
				br.close();
				fr.close();
				rc = true;
			} catch( IOException e ) {
				System.err.println(e);
			}
		}
		return rc;
	}
	
	public boolean saveFile( boolean bForceSave ) {
		boolean rc = false;
		if( m_file != null && ( isDirty() || bForceSave ) ) {
			try {
				FileWriter fw = new FileWriter(m_file);
				writeFile(fw);
				fw.close();
				rc = true;
			} catch( IOException e ) {
				System.err.println(e);
			}
		}
		return rc;
	}

///////////////////////////////////////////////////////////////////////////////
//  SplitScrollModel
    public void addView(SplitScrollView view) {
    	synchronized(m_viewlist) {
        	m_viewlist.addElement(view);
    	}
    }

    public int getColumnMaximum() {
        return 255;
    }

    public int getRowMaximum() {
        return Math.max(0, m_lineBuffer.size()-1);
    }

    public int getColumnMinimum() {
        return 0;
    }

    public int getRowMinimum() {
        return 0;
    }

    public void removeView(SplitScrollView view) {
    	synchronized(m_viewlist) {
	       m_viewlist.removeElement(view);
	    }
    }
    
    private void repaintViews() {
    	synchronized(m_viewlist) {
	    	Enumeration enum = m_viewlist.elements();
	    	while( enum.hasMoreElements() ) {
	    		SplitScrollView view = (SplitScrollView)enum.nextElement();
	    		view.repaint();
	    	}
	    	
	    }
    }
//  SplitScrollModel
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  DirtListener
	public int m_nDirtyLineCount = 0;
	public void dirtyChanged( boolean bDirty, DirtManager mgr ) {
		if( bDirty ) {
			m_nDirtyLineCount++;
			if( m_nDirtyLineCount == 1 ) {
				notifyDirty();
			}
		} else {
			m_nDirtyLineCount--;
			if( m_nDirtyLineCount == 0 ) {
				notifyDirty();
			}
		}
//		System.err.println("Dirty lines: "+m_nDirtyLineCount);
	}
	
	protected void notifyDirty() {
    	Iterator iter = m_dirtListeners.iterator();
    	while( iter.hasNext() ) {
    		DirtListener l = (DirtListener)iter.next();
    		l.dirtyChanged( isDirty(), this );
    	}
	}
//  DirtListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  DirtManager
    LinkedList m_dirtListeners = new LinkedList();
	public boolean isDirty() {
		return m_nDirtyLineCount > 0;
	}
	
	public void clearDirty() {
		for( int i=0; i<m_lineBuffer.size(); i++ ) {
			line(i).clearDirty();
		}
		m_nDirtyLineCount = 0;
	}
	
    public void addDirtListener( DirtListener l ) {
    	if( !m_dirtListeners.contains(l) ) {
    		m_dirtListeners.add(l);
    	} else {
    		System.err.println("Model already contains listener.");
    	}
    }
    
    public void removeDirtListener( DirtListener l ) {
    	m_dirtListeners.remove(l);
    }
//  DirtManager
///////////////////////////////////////////////////////////////////////////////
}