/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.project;

import java.util.*;
import java.io.*;
import gvid.DirtListener;
import gvid.util.*;
import gvid.*;

public class ProjectModel 
	implements	java.io.Serializable,
				gvid.DirtManager {
	// A project model is a list of files in a project, as well as 
	// language dependent settings.
	// The first version will be depth first and coded for the Java language
	// The project should be writeable and readable.
	
	// Unimplemented Stories:
	// Determine the version number of the jdkroot when set
	// jdk validate should use system dependent filenames 
	
	// Undefined Storied:
	// TODO: add more stories detailing root behavior relative to project root, and how they are updated if projec root changes.
	
	public ProjectModel(File projectFile) {
		this();
		setProjectFile( projectFile );
	}
	
	protected ProjectModel() {
		m_fileListListener = new ProjectListListener(this, m_fileList);
		m_classpathListListener = new ProjectListListener(this, m_classpath);	
	}
	
	// project window should display a name
	public String getName() {
		String rc = "new project";
		if( m_projectFile != null ) {
			rc = m_projectFile.getName();
		}
		return rc;
	}
	
	public File getProjectFile() {
		return m_projectFile;	
	}
	
	public void setProjectFile( File file ) {
		// srcRoot, dstRoot, docRoot, and jarRoot default to the location of the project file.
		// They all start out as null. When the project root is set, they are set if null.
		
		// TODO : someone should at least validate that the directories above the file exist
		if( !file.exists() || file.isFile() ) {
			m_projectFile = file;
		}
	}
	
	public String getMainClass() {
		return m_mainClass;
	}
	
	public String getArgs() {
		return m_args;
	}
	
	public int getReleaseType() {
		return m_releaseType;
	}
	
	public boolean getBuildDocs() {
		return m_buildJavaDoc;
	}
	
	public boolean getBuildJar() {
		return m_buildJar;
	}
	
	public File getSourceRoot() {
		File rc = m_srcRoot;
		if( rc == null ) {
			rc = new File(m_projectFile.getParent());
		}
		return rc;
	}
	
	public File getDestinationRoot() {
		File rc = m_dstRoot;
		if( rc == null ) {
			rc = new File(m_projectFile.getParent());
		}
		return rc;
	}
	
	public File getDocRoot() {
		File rc = m_docRoot;
		if( rc == null ) {
			rc = new File(m_projectFile.getParent());
		}
		return rc;
	}
	
	public File getJarRoot() {
		File rc = m_jarRoot;
		if( rc == null ) {
			rc = new File(m_projectFile.getParent());
		}
		return rc;
	}
	
	public void setMainClass( String mainClass ) {
		if( !mainClass.equals(m_mainClass) ) {
			m_mainClass = mainClass;
			setDirty();
		}
	}
	
	public void setArgs( String args ) {
		if( !args.equals(m_args) ) {
			m_args = args;
			setDirty();
		}
	}
	
	public void setReleaseType( int releaseType ) {
		if( releaseType!=m_releaseType ) {
			m_releaseType = releaseType;
			setDirty();
		}
	}
	
	public void setBuildDocs( boolean buildJavaDoc ) {
		if( buildJavaDoc!=m_buildJavaDoc ) {
			m_buildJavaDoc = buildJavaDoc;
			setDirty();
		}
	}
	
	public void setBuildJar( boolean buildJar ) {
		if( buildJar!=m_buildJar ) {
			m_buildJar = buildJar;
			setDirty();
		}
	}
	
	public void setSourceRoot( File file ) {
		if( file != null && !getSourceRoot().equals(file)) {
			if( !file.getAbsolutePath().equals( m_projectFile.getParent() ) ) {
				m_srcRoot = file;
			} else {
				m_srcRoot = null;
			}
			setDirty();
		}
	}
	
	public void setDestinationRoot( File file ) {
		if( file != null && !getDestinationRoot().equals(file)) {
			if( !file.getAbsolutePath().equals( m_projectFile.getParent() ) ) {
				m_dstRoot = file;
			} else {
				m_dstRoot = null;
			}
			setDirty();
		}
	}
	
	public void setDocRoot( File file ) {
		if( file != null && !getDocRoot().equals(file)) {
			if( !file.getAbsolutePath().equals( m_projectFile.getParent() ) ) {
				m_docRoot = file;
			} else {
				m_docRoot = null;
			}
			setDirty();
		}
	}
	
	public void setJarRoot( File file ) {
		if( file != null && !getJarRoot().equals(file)) {
			if( !file.getAbsolutePath().equals( m_projectFile.getParent() ) ) {
				m_jarRoot = file;
			} else {
				m_jarRoot = null;
			}
			setDirty();
		}
	}
	
	// A way to access files in the project.
	public List getFileList() {
		NotifyingList list = new NotifyingList( m_fileList );
		// This references is one way so the NotifyingList should 
		// be garbage collected once the caller is done with it.
		list.addListListener(m_fileListListener);
		return list;
	}
	
	public List getClasspath() {
		NotifyingList list = new NotifyingList( m_classpath );
		// This references is one way so the NotifyingList should 
		// be garbage collected once the caller is done with it.
		list.addListListener(m_classpathListListener);
		return list;
	}
	
	public int getFileCount() {
		return m_fileList.size();
	}
	
	public File getJDKRoot() {
		File rc = null;
		rc = m_jdkRoot;
		return rc;
	}
	
	// a way to set the jdk root
	// the jdkroot is validated when set
	// it fails if this is not a valid jdk
	public boolean setJdkRoot( File root ) {
		boolean rc = validateJdkRoot(root);
		if( rc && !root.equals(m_jdkRoot) ) {
			m_jdkRoot = new File( root.getAbsolutePath() );
			setDirty();
		}
		return rc;
	}
	
	// verify that the jdkroot is a jdkroot
	public boolean validateJdkRoot( File root ) {
		// look for java, javac, jdb, jar, javadoc
		boolean rc = true;
		
		rc = rc && root.isDirectory();		
		
		root = new File( root, "bin" );
		
		if( rc ) rc = rc && new File(root, "java.exe" ).exists();
		if( rc ) rc = rc && new File(root, "javac.exe" ).exists();
		if( rc ) rc = rc && new File(root, "jdb.exe" ).exists();
		if( rc ) rc = rc && new File(root, "jar.exe" ).exists();
		if( rc ) rc = rc && new File(root, "javadoc.exe" ).exists();
		
		return rc;
	}
	
	// A root for the file sources 
	private File m_srcRoot = null;
	
	// A directory in which to put the class files 
	private File m_dstRoot = null;
	
	// A directory in which to put the documentation 
	private File m_docRoot = null;
	
	// A directory in which to put the jar files 
	private File m_jarRoot = null;
	
	// A jdk installation location. Used to get all tools (java, javac, jdb, javadoc)
	private File m_jdkRoot = null;
	
	// A list of additional class files for compiling and running. (classpath)
	private Vector m_classpath = new Vector();
	
	// A list of files in that location that are included in the project.
	private Vector m_fileList = new Vector();
	
	// Debug/final option
	// Program arguments option
	public static final int DEBUG = 0;
	public static final int FINAL = 1;
	private int m_releaseType = DEBUG;
	
	// Build javadoc option
	private boolean m_buildJavaDoc = false;
	
	// Build jar option
	private boolean m_buildJar = false;
	
	// Main class option
	private String m_mainClass = "";
	
	// Program arguments option
	private String m_args = "";

	// A project filename
	transient private File m_projectFile = null;
	
	transient private ProjectListListener m_fileListListener = null;
	transient private ProjectListListener m_classpathListListener = null;
	transient private DirtManagerImpl m_dirtImpl = new DirtManagerImpl(this);
	transient private Vector m_projectListeners = new Vector();
	
	static final long serialVersionUID = 6376629515675225515L;
	private void readObject(ObjectInputStream in) 
		throws	IOException, 
				ClassNotFoundException {
		in.defaultReadObject();
		
		m_dirtImpl = new DirtManagerImpl(this);
		m_fileListListener = new ProjectListListener(this, m_fileList);
		m_classpathListListener = new ProjectListListener(this, m_classpath);
		m_projectListeners = new Vector();
		clearDirty();
	}
///////////////////////////////////////////////////////////////////////////////
// DirtManager implementation
	public boolean isDirty() { 
		boolean rc = true;
		if( m_dirtImpl != null ) {
			rc = m_dirtImpl.isDirty();
		}
		return rc;
	}

	public void clearDirty() {
		if( m_dirtImpl != null ) {
			m_dirtImpl.clearDirty();
		}
		if( m_fileListListener != null ) {
			m_fileListListener.clear();
		}
		if( m_classpathListListener != null ) {
			m_classpathListListener.clear();
		}
	}
	
	public void addDirtListener( DirtListener l ) {
		if( m_dirtImpl != null ) {
			m_dirtImpl.addDirtListener(l);
		}
	}
	
	public void removeDirtListener( DirtListener l ) {
		if( m_dirtImpl != null ) {
			m_dirtImpl.removeDirtListener(l);
		}
	}
	
	void updateDirty() {
		if( m_fileListListener.isDirty() || m_classpathListListener.isDirty() ) {
			setDirty();
		} else {
			clearDirty();
		}
	}
	
	private void setDirty() {
		if( m_dirtImpl != null ) {
			m_dirtImpl.setDirty();
		}
	}
// DirtManager implementation
///////////////////////////////////////////////////////////////////////////////
	
	public void addProjectListener( ProjectListener l ) {
		m_projectListeners.add(l);
	}
	
	public void removeProjectListener( ProjectListener l ) {
		m_projectListeners.remove(l);
	}
	
	void notifyFilesChanged() {
		Iterator iter = m_projectListeners.iterator();
		while( iter.hasNext() ) {
			ProjectListener l = (ProjectListener)iter.next();
			l.filesChanged();
		}
	}
}
