/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.ui.editor;

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.dnd.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import gvid.command.*;
import gvid.document.*;
import gvid.ui.splitscrollpane.*;
import gvid.ui.editor.undo.*;
import gvid.ui.editor.selection.*;
import gvid.util.*;
import gvid.undo.*;
import gvid.*;
import gvid.ui.status.*;

/**
 * This class provides a way for viewing and editing source code. It differs from a standard 
 * editor in the following ways:
 * - Editing font is not proportional. Every character has the same width
 * - No line or word wrapping
 * - Each line has a maximum length of 255
 * 
 * This class is defined to be used in a SplitScrollPane. The document model can be shared 
 * between multiple views.
 * 
 * @author Alan Oursland
 * @version 1.00
 * @since 2/26/2001
 * @see SplitScrollPane
 * @see SourceCodeModel
 * @see EditCaret
 */
public class SourceCodeView 
	extends SplitScrollViewBase
	implements SourceCodeEditor,
			   FocusListener,
			   MouseListener,
			   MouseMotionListener,
			   KeyListener,
			   ChangeListener,
			   LayoutManager,
			   DropTargetListener,
			   ContainerListener,
			   Autoscroll,
			   DirtListener,
			   DirtManager,
			   StatusSource,
			   EditCaretListener
{
//	String m_tabDisplay = INVISIBLETAB;
	private SourceCodeModel 		m_sourceModel			= null;
	private SourceCodeEditBar		m_bar					= null;
	private DropTarget				m_target				= null;
	private EditCaret				m_caret 				= null;
	private SourceCodeSelection 	m_selection 			= null;
	private int						m_selectionMode 		= SourceCodeSelection.BOUNDED;
	private AutoscrollController	m_autoscrollController	= null;
	private boolean 				m_bDragSelectionMode	= false;
	private Point					m_originalCaretPosition	= new Point();
    private boolean 				m_bDelayCancelSelection	= false;
	private boolean 				m_bSelectWordMode		= false;
	private Point					m_selectWordFrontAnchor	= null;
	private Point					m_selectWordBackAnchor	= null;
	private Vector					m_viewReferences		= null;
	private EditorEditCommandsImpl	m_cmdImpl				= null;
	private boolean 				m_bHasFocus 			= false;
    private LinkedList				m_dirtListeners 		= null;
	private String					m_statusMsg 			= "Ready";
	private Vector					m_statusListeners		= null;
	private String[]				m_statusStr 			= {
		"Ln 1, Col 1   ",
		"REC",
		"COL",
		"OVR",
		"READ"
	};
	private boolean[]				m_statusActive			= {
		true,
		false,
		false,
		false,
		false
	};

	public SourceCodeView() {
		setFont(new Font("Courier New", Font.PLAIN, 14));
		buildCursor(this);
//		m_sourceModel				= new SourceCodeModel();
		m_selectionMode 			= SourceCodeSelection.BOUNDED;
		m_autoscrollController		= new AutoscrollController(this);
		m_originalCaretPosition 	= new Point();
		m_viewReferences			= new Vector();
		m_cmdImpl					= new EditorEditCommandsImpl(this);
    	m_dirtListeners 			= new LinkedList();
		m_statusListeners			= new Vector();
		m_caret 					= new EditCaret(this);		
		m_bar						= new SourceCodeEditBar(this);

		setModel(new SourceCodeModel());
		
		m_target					= new DropTarget(this, this);
		gvid.Util.recurseCreateDropTarget(this, this, new gvid.Util.RecurseDropTargetComponentListener(this));
		
		setLayout(this);
		setOpaque(false);
		setBackground(Color.white);
		setForeground(Color.black);
		
		setCursor( Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
		add(m_bar);

		addFocusListener(this);
		addMouseListener(this);
		addMouseMotionListener(this);
		addKeyListener(this);
		addContainerListener(this);
		m_caret.addMouseListener(this);
		m_caret.addMouseMotionListener(this);
		m_caret.addCaretListener(this);
	}

	public void addNotify() {
		super.addNotify();
		getHorizontalScrollModel().addChangeListener(this);
		getVerticalScrollModel().addChangeListener(this);		 
	}

	public void removeNotify() {
		super.removeNotify();
		getHorizontalScrollModel().removeChangeListener(this);
		getVerticalScrollModel().removeChangeListener(this);		
	}

	public void doLayout() {
		super.doLayout();
		m_caret.updateBounds();
		m_bar.updateBounds();
	}

	public void loadFile( String filename ) {
		try {
			setModel(new SourceCodeModel());
			FileReader fr = new FileReader(filename);
			BufferedReader br = new BufferedReader(fr);
			String currentLine = br.readLine();
			while( currentLine != null ) {
				getSourceModel().addLine(new Line(currentLine));
				currentLine = br.readLine();
			}
			getSourceModel().clearDirty();
			br.close();
			fr.close();
		} catch( IOException e ) {
			System.err.println(e);
		}
	}

	public void paint( Graphics g ) {
		g.setColor(Color.white);
		g.setClip(0, 0, getBounds().width, getBounds().height);
		g.fillRect(0, 0, getBounds().width, getBounds().height);

		g.setColor(Color.black);
		FontMetrics fm = g.getFontMetrics();
//		int yPos = fm.getAscent();
		for( int i=getFirstVisibleRow(); i<=getLastVisibleRow()+1; i++ ) {
			Point pos = new Point(getFirstVisibleColumn(), i);
			Point doc = posToDoc(pos);
			int tabLocation = doc.x;
			tabLocation = getSourceModel().lineIndexOf(i, '\t', tabLocation);
			while( tabLocation != -1 ) {
				Point view = positionToView(docToPos(doc));
				g.drawString( getSourceModel().lineToString(i).substring(doc.x, tabLocation), view.x, view.y+fm.getAscent());
				doc.x = tabLocation;
				view = positionToView(docToPos(doc));
				g.drawString( gvid.Globals.getTabDisplay(), view.x, view.y);
				doc.x++;
				tabLocation = getSourceModel().lineIndexOf(i, '\t', doc.x);
			}
			if( doc.x < getLineLength(i) ) {
				Point view = positionToView(docToPos(doc));
				g.drawString(getSourceModel().lineToString(i).substring(doc.x), view.x, view.y+fm.getAscent());
			}
			
			if( gvid.Globals.getTabDisplay() == gvid.Globals.VISIBLETAB ) {
			    // draw dots for each space
			    g.setColor(Color.black);
			    // we need to draw dots in all of the spaces
			    Point d = new Point(getSourceModel().lineIndexOf(doc.y, ' ',0), doc.y);
			    while( d.x != -1 ) {
			        Point view = positionToView(docToPos(d));
			        g.drawLine( view.x+fm.charWidth(' ')/3, view.y+fm.getAscent()/2,view.x+fm.charWidth(' ')/3, view.y+fm.getAscent()/2);
			        d.x = getSourceModel().lineIndexOf(i, ' ',d.x+1);
			    }
			}
			
		}
		super.paint(g);	   
	}
	
	public Dimension getDocumentSize() {
		Dimension rc = new Dimension();

		FontMetrics fm = getFontMetrics(getFont());
		rc.height = (getLineCount()+1)*getFontHeight();
		rc.width = 255*fm.charWidth(' ');

		return rc;
	}	
	
	public boolean isManagingFocus() {
		return true;
	}
	
	public StatusSource getStatusSource() {
		return this;
	}
	
///////////////////////////////////////////////////////////////////////////////
//	Overridden methods
    protected int getExtent(int alignment) {
        int rc = 1;
        if( alignment == SwingConstants.HORIZONTAL ) {
            rc = horizontalUnits(getEditBounds().width-getFontWidth()/2);
        } else {
            rc = verticalUnits(getEditBounds().height-getFontHeight()/2);
        }
        return rc;
    }
//	Overridden methods
///////////////////////////////////////////////////////////////////////////////
    
	
///////////////////////////////////////////////////////////////////////////////
//	SplitScrollView
	public SplitScrollView createCopy() {
		SplitScrollView rc = new SourceCodeView();
		rc.copySettingsFrom(this);
		return rc;
	}

	public void copySettingsFrom(SplitScrollView view) {
		if( view != this ) {
			super.copySettingsFrom(view);
			if( view instanceof SourceCodeView ) {
//				setModel(((SourceCodeView)view).m_sourceModel);
				m_caret.copySettingsFrom(((SourceCodeView)view).m_caret);
				m_bar.copySettingsFrom(((SourceCodeView)view).m_bar);
				if( ((SourceCodeView)view).m_selection != null ) {
					m_selection = ((SourceCodeView)view).m_selection.createCopy(this);
					add(m_selection.asComponent(), 0);
				} else {
					if( m_selection != null ) {
						remove(m_selection.asComponent());
						repaint(m_selection.asComponent().getBounds());
					}
					m_selection = null;
				}
			}
		}
	}

	public void clearModel() {
		super.clearModel();
		if( m_sourceModel != null ) {
			m_sourceModel.removeDirtListener(this);
			// caret must be removed so it can remove the reference point
			// it will be re-added when a new model is added
			remove(m_caret);
			repaint(m_caret.getBounds());
			if( m_selection != null ) {
				clearSelection();
			}
			m_sourceModel = null;
		}
		// this must happen before the model is changed.
		
	}

	public void setModel(SplitScrollModel model) {
		clearModel();
		if( ( model != null ) && 
		    ( model != m_sourceModel) &&
		    ( model instanceof SourceCodeModel ) ) {

			super.setModel(model);

			m_sourceModel = (SourceCodeModel)model;
			m_sourceModel.addDirtListener(this);
			
			// why is the caret set non visible?
			m_caret.setVisible(false);
			
			// re-add the caret so it adds the reference position
			// 99% of the time in this function happens here
			add(m_caret);
			
			notifyDirty();
		}
	}

	public int horizontalUnits( int width ) {
		return width/getFontWidth();
	}

	public int verticalUnits( int height ) {
		return height/getFontHeight();
	}

//	SplitScrollView
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	SourceCodeEditor
	public UndoRedoManager getUndoManager() {
		return m_sourceModel.getUndoManager();
	}

	public int getFontHeight() {
		FontMetrics fm = getFontMetrics(getFont());
		return fm.getAscent()+fm.getDescent();
	}

	public int getFontWidth() {
		return getFontMetrics().charWidth(' ');
	}

	public FontMetrics getFontMetrics() {
		return getFontMetrics(getFont());
	}

	public int getFirstVisibleColumn() {
		return getHorizontalScrollModel().getValue();
	}

	public int getFirstVisibleRow() {
		return getVerticalScrollModel().getValue();
	}	 

    public int getLastVisibleColumn() {
    	return getHorizontalScrollModel().getValue()+getHorizontalScrollModel().getExtent();
    }
    
    public void setFirstVisibleRow( int row ) {
    	getVerticalScrollModel().setValue(row);
    }
    
    public void setFirstVisibleColumn( int col ) {
    	getHorizontalScrollModel().setValue(col);
    }
    
    public int getLastVisibleRow() {
    	return Math.min( 
    		getLineCount(), 
    		getFirstVisibleRow() + (getEditBounds().height+1)/getFontHeight()-1);
    }
    
    public int getVisibleRowCount() {
    	return getVerticalScrollModel().getExtent();
    }
    
    public int getVisibleColumnCount() {
    	return getHorizontalScrollModel().getExtent();
    }
    
	public int getColumnLeftEdge(int col) {
		return m_bar.getBounds().width + (col-getFirstVisibleColumn())*getFontWidth();
	}

	public int getRowTopEdge(int row) {
		return (row-getFirstVisibleRow())*getFontHeight();
	}

	public int getTabWidth() {
		return 4;	
	}

    public boolean isInsert() {
    	return EditCaret.isInsert();
    }

    public boolean isOverwrite() {
    	return EditCaret.isOverwrite();
    }
    
	/**
	 * Converts view pixel based coordinates into row/column screen coordinates.
	 * 
	 * @param viewCoord
	 */
	public Point viewToPosition( Point viewCoord ) {
		Point rc = new Point();
		rc.y = (viewCoord.y-getEditBounds().y)/getFontHeight()+getFirstVisibleRow();
		rc.x = (viewCoord.x-getEditBounds().x)/getFontWidth()+getFirstVisibleColumn();
		return rc;
	}

	/**
	 * Converts view row/column position into view pixel coordinates.
	 * 
	 * @param position
	 */
	public Point positionToView( Point position ) {
		Point rc = new Point();
		rc.y = (position.y-getFirstVisibleRow())*getFontHeight()+getEditBounds().y;
		rc.x = (position.x-getFirstVisibleColumn())*getFontWidth()+getEditBounds().x;
		return rc;
	}

	/**
	 * Converts view row/column position into document row/column position.
	 * 
	 * @param position
	 */
	public Point posToDoc( Point position ) {
		Point rc = new Point(position);

		if( getSourceModel() != null ) {
			// Make sure row is valid
			rc.y = Math.max(rc.y, getFirstLine());
			rc.y = Math.min(rc.y, getLineCount());

			// adjust columns for tabs
			// i is document position
			// p is view position
			int p = 0;
			for( int i=0; i<rc.x && i<getLineLength(rc.y); i++ ) {
				if( getSourceModel().charAt(new Point(i, rc.y)) == '\t' ) {
					if( rc.x < i+getTabWidth() - p%getTabWidth() ) {
						rc.x = i;
					} else {
						rc.x -= getTabWidth() - p%getTabWidth() - 1;
						p += getTabWidth() - p%getTabWidth() - 1;
						// the mod operation ensures that tabs aren't just 4 spaces
						// they should go to every 4th position
					}
				}
				p++;
			}

			// make sure column is valid
			rc.x = Math.max(rc.x, getSourceModel().getColumnMinimum());		  
			rc.x = Math.min(rc.x, getLineLength(rc.y));
		} else {
			rc.x = 0;
			rc.y = 0;
		}

		return rc;
	}

	/**
	 * Converts document row/column position to view row/column position.
	 * 
	 * @param doc
	 */
	public Point docToPos( Point doc ) {
		Point rc = new Point(doc);
		// Make sure row is valid
		rc.y = Math.max(rc.y, getFirstLine());
		rc.y = Math.min(rc.y, getLineCount());
		// make sure column is valid
		rc.x = Math.max(rc.x, getSourceModel().getColumnMinimum());		  
		rc.x = Math.min(rc.x, getLineLength(rc.y));

		// adjust columns for tabs
		// i is document position
		// p is view position
		int p = 0;
		for( int i=0; i<doc.x && i<getLineLength(rc.y); i++ ) {
			if( getSourceModel().charAt(new Point(i, rc.y)) == '\t' ) {
				rc.x += getTabWidth()-1 - p%getTabWidth();	
				p += getTabWidth()-1 - p%getTabWidth();
					// the mod operation ensures that tabs aren't just 4 spaces
					// they should go to every 4th position
			}
			p++;
		}

		return rc;
	}

	public Point getCaretPosition() {
		return m_caret.getCaretPosition();
	}

	/**
	 * Sets the caret location using a view position. The caret remembers the view 
	 * position but converts it into a document position for actual placement.
	 * 
	 * @param col
	 * @param row
	 */
	public UndoCommand setCaretPosition( Point pos, boolean bClearSelection ) {
		UndoChain rc = new UndoChain("setCaretPosition");
		if( bClearSelection && getSelection() != null ) {
			rc.addRedo(clearSelection());
		}
		Point oldPos = m_caret.getCaretPosition();
		m_caret.setCaretPosition(pos);
		if( !pos.equals(oldPos) ) {
			rc.addRedo( new MoveCaretCommand( pos.x, pos.y, oldPos.x, oldPos.y, "setCaretPosition" ) );
		}
		return rc;
	}

	public void scrollToCaret() {
		Point p = m_caret.getCaretPosition();
		Point d = posToDoc(p);
		p = docToPos(d);
		if( p.x < getFirstVisibleColumn() ) {
			getHorizontalScrollModel().setValue(p.x-4);
			repaint();
		}
		if( p.y < getFirstVisibleRow() ) {
			getVerticalScrollModel().setValue(p.y);
			repaint();
		}
		if( p.x > getFirstVisibleColumn()+getHorizontalScrollModel().getExtent() ) {
			getHorizontalScrollModel().setValue(p.x-getHorizontalScrollModel().getExtent()+4);
			repaint();
		}
		if( p.y > getFirstVisibleRow()+getVerticalScrollModel().getExtent() ) {
			getVerticalScrollModel().setValue(p.y-getVerticalScrollModel().getExtent());
			repaint();
		}
	}
	
	public UndoCommand setSelectionMode( int mode ) {
		UndoCommand rc = null;
		if( m_selectionMode != mode ) {
			rc = new SelectionModeCommand("setSelectionMode", mode, m_selectionMode);
			m_selectionMode = mode;
		}
		if( mode == SourceCodeSelection.BLOCKED ) {
			setColumnStatus(true);
		} else {
			setColumnStatus(false);
		}
		
		return rc;
	}

    public UndoCommand setSelectionPosition( Point selectionPos ) {
    	SelectionCommand rc = new SelectionCommand( "setSelectionCommand" );
    	rc.initSelection(this);
    	if( m_selection != null && m_selection.getSelectionType() == m_selectionMode ) {
    		m_selection.setSelectionEnd(selectionPos);
    	} else {
    		switch( m_selectionMode ) {
    			case SourceCodeSelection.BOUNDED: {
					if( m_selection != null ) {
						remove(m_selection.asComponent());
						repaint(m_selection.asComponent().getBounds());
    					m_selection = new BoundedSelection(m_selection.getSelectionStart(), selectionPos, this);
					} else {
    					m_selection = new BoundedSelection(getCaretPosition(), selectionPos, this);
					}
    				add(m_selection.asComponent(), 0);
    			} break;
    			case SourceCodeSelection.BLOCKED: {
					if( m_selection != null ) {
						remove(m_selection.asComponent());
						repaint(m_selection.asComponent().getBounds());
    					m_selection = new BlockSelection(m_selection.getSelectionStart(), selectionPos, this);
					} else {
    					m_selection = new BlockSelection(getCaretPosition(), selectionPos, this);
					}
    				add(m_selection.asComponent(), 0);
    			} break;
    			case SourceCodeSelection.LINE: {
					if( m_selection != null ) {
						remove(m_selection.asComponent());
						repaint(m_selection.asComponent().getBounds());
						// TODO : we need to convert BoundedSelection into LineSelection better
    					m_selection = new LineSelection(m_selection.getSelectionStart(), selectionPos, this);
					} else {
    					m_selection = new LineSelection(getCaretPosition(), selectionPos, this);
					}
    				add(m_selection.asComponent(), 0);
    			} break;
    		}
    	}
    	setCaretPosition(selectionPos, false);
    	// this is not undone - selSelectionPosition is handling its own undo
    	invalidate();
    	
    	if( m_selection != null ) {
    		if( m_selection.getSelectionStart().equals(m_selection.getSelectionEnd())) {
    			clearSelection();
    			// this is not undone - selSelectionPosition is handling its own undo
    		}
    	}
    	rc.finalSelection(this);
    	return rc;
    }
    
    public SourceCodeSelection getSelection() {
    	return m_selection;
    }
    
    public UndoCommand clearSelection() {
    	SelectionCommand rc = new SelectionCommand("clearSelection");
    	rc.initSelection(this);
    	if( m_selection != null ) {
    		remove(m_selection.asComponent());
			repaint(m_selection.asComponent().getBounds());
    		m_selection = null;
    		repaint();
    	}
    	rc.finalSelection(this);
    	return rc;
    }
    
	public SourceCodeModel getSourceModel() {
		return m_sourceModel;
	}	
	
	public int getLineLength( int row ) {
		return getSourceModel().lineLength(row);
	}
	
    public int getLineCount() {
    	return getSourceModel().getRowMaximum();
    }	
    
    public int getFirstLine() {
    	return getSourceModel().getRowMinimum();
    }
	
	public Rectangle getEditBounds() {
		Rectangle r = getBounds();
		r.x = m_bar.getWidth();
		r.y = 0;
		r.width -= m_bar.getWidth();
		return r;
	}
	
    public UndoCommand insertChar( char c ) {
		UndoChain rc = new UndoChain("insertChar");
    	if( getSelection() != null ) {
    		rc.addRedo(getEditCommands().delete());
    	}
		Point p = getCaretPosition();
		Point d = posToDoc(p);		
		p = docToPos(d);
		rc.addRedo(setCaretPosition(p, true));
    	rc.addRedo(getSourceModel().insertChar(c, d));
    	return rc;
    }
    
    public UndoCommand overwriteChar( char c ) {
		UndoChain rc = new UndoChain("overwriteChar");
		
    	if( getSelection() != null ) {
    		rc.addRedo(getEditCommands().delete());

			Point p = getCaretPosition();
			Point d = posToDoc(p);		
			p = docToPos(d);
			
			rc.addRedo(setCaretPosition(p, true));
    		rc.addRedo(getSourceModel().insertChar(c, d));
    	} else {
			Point p = getCaretPosition();
			Point d = posToDoc(p);		
			p = docToPos(d);
			
			rc.addRedo(setCaretPosition(p, true));
    		rc.addRedo(getSourceModel().overwriteChar(c, d));
    	}
		
    	return rc;   	
    }
    
    public UndoCommand deleteRange( Point startPos, Point endPos ) {
    	Point docStart = posToDoc(startPos);
    	Point docEnd = posToDoc(endPos);
    	return getSourceModel().deleteRange(docStart, docEnd);
    }
    
    public DataFlavor getPreferredDataFlavor( DataFlavor[] df ) {
    	DataFlavor rc = null;
        for( int i=0; i < df.length; i++ ) {
        	if( df[i].getPrimaryType().equals("text") && df[i].getSubType().equals("plain") ) {
        		if( rc == null ) {
        			rc = df[i];
        		}
        	} else if( SourceCodeSelection.class.isAssignableFrom(df[i].getRepresentationClass()) ) {
        		rc = df[i];
        		break;
        	}
        }
        return rc;
    }
    
    public Object getPreferredTransferableObject( Transferable t, DataFlavor f ) {
    	Object rc = null;
    	try {
    		if( SourceCodeSelection.class.isAssignableFrom(f.getRepresentationClass()) ) {
    			rc = t.getTransferData(f);
    		} else if( f.getPrimaryType().equals("text") && f.getSubType().equals("plain") ) {
    			rc = f.getReaderForText(t);
    		} else {
    			rc = t.getTransferData(f);
    		}
    	} catch( UnsupportedFlavorException e ) {
    		System.err.println(e);
    	} catch( IOException e ) {
    		System.err.println(e);
    	}
    	return rc;
    }
    
    public UndoCommand insertTransferableObject( Object obj, DataFlavor flavor ) {
    	UndoChain rc = new UndoChain("insertTransferableObject");
     	if( obj != null ) {
    		if( obj instanceof SourceCodeSelection ) {
    			SourceCodeSelection sel = (SourceCodeSelection)obj;
    			rc.addRedo(sel.paste(this, getCaretPosition()));
    		} else if( obj instanceof Reader ) {
    			try {
        			Reader in = (Reader)obj;
    				StringBuffer str = new StringBuffer();
    				int c = in.read();
    				while( c != -1 && c != 0) {
    					str.append((char)c);
    					c = in.read();
    				}
    				StringBuffer extraInfo = new StringBuffer();
    				while( c != -1 ) {
    					extraInfo.append((char)c);
    					c = in.read();
    				}
    				if( extraInfo.length() > 0 ) {
    					System.err.print("Extra transferrable info:");
    					for( int j=0; j<extraInfo.length(); j++ ) {
    						System.err.print("\t"+(int)extraInfo.charAt(j));
    					}
    					System.err.println();
    				}
    						
    				in.close();
    				rc.addRedo(getSourceModel().insertText(str.toString(), posToDoc(getCaretPosition())));
    			} catch( IOException e ) {
    				System.err.println(e);
    			}
    		} else {
    			System.err.println(obj.toString());
        	}
    	}
    	return rc;
    }
    
    public String getText( Point startPos, Point endPos ) {
    	Point docStart = posToDoc(startPos);
    	Point docEnd = posToDoc(endPos);
    	
    	return getSourceModel().getText( docStart, docEnd );
    }
    
    public void startSelectionDrag(DropTargetDragEvent dtde) {
    	if( !m_bDragSelectionMode ) {
        	DataFlavor[] df = dtde.getCurrentDataFlavors();
    		DataFlavor preferredFlavor = getPreferredDataFlavor(df);
        	if( preferredFlavor != null ) {
        		if( (dtde.getDropAction()&DnDConstants.ACTION_COPY_OR_MOVE) != 0 ) {
					dtde.acceptDrag(dtde.getDropAction());
        		}
				m_caret.setMovingMode(true);
				m_caret.setVisible(true);
				m_originalCaretPosition = getCaretPosition();
				m_bDragSelectionMode = true;
				updateSelectionDrag(dtde);
        	} else {
        		dtde.rejectDrag();
        	}
    	} else {
    		synchronized(this) {
    			m_bDelayCancelSelection = false;
    			notify();
    		}
    	}
    }
    
    public void cancelSelectionDrag() {
    	m_bDelayCancelSelection = true;
    	new Thread() {
    		public void run() {
    			synchronized(this) {
    				try {
    					wait(500);    			
    				} catch( java.lang.InterruptedException e ) {
    					System.err.println(e);
    				}
    			}
    			if( m_bDelayCancelSelection && m_bDragSelectionMode ) {
    				SwingUtilities.invokeLater(new Runnable() {
    					public void run() {
							m_caret.setMovingMode(false);
							m_caret.setVisible(m_bHasFocus);
							m_bDragSelectionMode = false;
    						setCaretPosition(m_originalCaretPosition, false);
    						// This command does not undo since it is restoring to the 
    						// previous state
    					}
    				});
    			}
    		}
    	}.start();
    }
    
    public void completeSelectionDrag( DropTargetDropEvent dtde ) {
    	synchronized(this) {
    		m_bDelayCancelSelection = false;
    		notify();
    	}
    	if( m_bDragSelectionMode ) {
    		if( m_caret.isVisible() ) {
        		DataFlavor[] df = dtde.getCurrentDataFlavors();
    			final DataFlavor preferredFlavor = getPreferredDataFlavor(df);
        		if( preferredFlavor != null ) {
        			if( (dtde.getDropAction()&DnDConstants.ACTION_COPY_OR_MOVE) != 0 ) {
						dtde.acceptDrop(dtde.getDropAction());
        			}
        			Transferable trans = dtde.getTransferable();
        			final Object data = getPreferredTransferableObject(trans, preferredFlavor);
					SwingUtilities.invokeLater(
						new Runnable() {
							public void run() {
								DragAndDropCommand undoCmd = new DragAndDropCommand("dragAndDrop", data);
								// creating this undo command with the data allows the drag and drop commands to be combined if necessary.
								Point insertionStart = docToPos(posToDoc(getCaretPosition()));
								undoCmd.addRedo(insertTransferableObject(data, preferredFlavor));
								Point insertionEnd = getCaretPosition();
									
								// we need to set the selection like this so that the original
								// selection can finish the drop
								// the original will remove itself after the drop
								UndoCommand modeCmd = null;
								UndoCommand selCmd = null;
								if( data instanceof BlockSelection ) {
									setCaretPosition(insertionStart, true);
									modeCmd = setSelectionMode(SourceCodeSelection.BLOCKED);
									selCmd = setSelectionPosition(insertionEnd);
									// undo handled by SelectionCommand
								} else {
									setCaretPosition(insertionStart, true);
									modeCmd = setSelectionMode(SourceCodeSelection.BOUNDED);
									selCmd = setSelectionPosition(insertionEnd);
									// undo handled by SelectionCommand
								}
								undoCmd.addRedo(modeCmd);
								undoCmd.addRedo(selCmd);
    							if( getSelection() != null ) {
    								getSelection().asComponent().setVisible(m_bHasFocus);
    							}
//								System.err.println(undoCmd.getId());
								getUndoManager().addUndo(undoCmd);
							}
						});
						dtde.dropComplete(true);
        			} else {
						dtde.dropComplete(false);
        				getUndoManager().addUndo(clearSelection());
        			}
    		}
			m_caret.setVisible(m_bHasFocus);
			m_caret.setMovingMode(false);
			m_bDragSelectionMode = false;
    	}
    }
    
    public void updateSelectionDrag( DropTargetDragEvent dtde ) {
     	synchronized(this) {
    		m_bDelayCancelSelection = false;
    		notify();
    	}
	   	if( m_bDragSelectionMode ) {
    		Point view = dtde.getLocation();
    		view = SwingUtilities.convertPoint(dtde.getDropTargetContext().getComponent(), view, this);    		
    		view.x += getFontWidth()/2;
    		Point pos = viewToPosition(view);
    		if( ( getSelection() != null ) && 
    		    ( getSelection().asComponent().isVisible()) && 
    		    ( getSelection().containsPos(pos) ) ) {
    		    // if there is a selection, and the selection is visible (meaning
    		    // the editor has the focus), and the selection contains the
    		    // mouse location, hide the caret. This turns the caret off when
    		    // dragging a selection through itself.
    			m_caret.setVisible(false);
    		} else {
    			m_caret.setVisible(true);
    			setCaretPosition(pos, false);
    			// this does not undo. It is only used as an indicator for dropping a selection
    		}
    	}
    }
    
    public UndoCommand startWordSelection( Point view ) {
		m_bSelectWordMode = true;

    	Point pos = viewToPosition(view);
    	Point doc = posToDoc(pos);
    	m_selectWordFrontAnchor = getSourceModel().wordStart(doc);
    	m_selectWordBackAnchor = getSourceModel().wordEnd(doc);
    	Point selStart = docToPos(m_selectWordFrontAnchor);
    	Point selEnd = docToPos(m_selectWordBackAnchor);
    	
    	UndoChain rc = new UndoChain("startWordSelection");
    	setCaretPosition(selStart, true);
    	rc.addRedo( setSelectionMode( SourceCodeSelection.BOUNDED ) );
    	rc.addRedo( setSelectionPosition(selEnd) );
		return rc;
    }
    
    public UndoCommand stopWordSelection( Point view ) {
    	UndoCommand rc = updateWordSelection(view);
		m_bSelectWordMode = false;		
		return rc;
    }
    
    public UndoCommand updateWordSelection( Point view ) {
    	Point pos = viewToPosition(view);
    	Point doc = posToDoc(pos);
    	
    	UndoChain rc = new UndoChain("updateWordSelection");
    	if( isBefore(doc, m_selectWordFrontAnchor) ) {
    		// make sure the back is the anchor
    		Point p = docToPos(m_selectWordBackAnchor);
    		setCaretPosition(p, true);
    		doc = getSourceModel().wordStart(doc);
    		rc.addRedo(setSelectionMode(SourceCodeSelection.BOUNDED));
    		rc.addRedo(setSelectionPosition(docToPos(doc)));
    	} else if( isBefore(m_selectWordBackAnchor, doc) ) {
    		// make sure the front is the anchor
    		Point p = docToPos(m_selectWordFrontAnchor);
    		setCaretPosition(p, true);
    		doc = getSourceModel().wordEnd(doc);
    		rc.addRedo(setSelectionMode(SourceCodeSelection.BOUNDED));
    		rc.addRedo(setSelectionPosition(docToPos(doc)));
    	} else {
    		Point p = docToPos(m_selectWordFrontAnchor);
    		setCaretPosition(p, true);
    		rc.addRedo(setSelectionMode(SourceCodeSelection.BOUNDED));
    		rc.addRedo(setSelectionPosition(docToPos(m_selectWordBackAnchor)));
    	}
    	scrollToCaret();
    	repaint();    	
    	return rc;
    }
    
    public EditCommands getEditCommands() {
    	return m_cmdImpl;
    }
    
    public boolean addReferencePoint(Point p, String name, ReferencePointListener l) {
    	boolean rc = false;
    	if( getSourceModel().addReferencePoint(p, name, l) ) {
    		m_viewReferences.addElement(name);
    		rc = true;
    	}
    	return rc;
    }
    
    public void removeReferencePoint(String name, ReferencePointListener l) {
    	m_viewReferences.remove(name);
    	if( getSourceModel() != null ) {
    		getSourceModel().removeReferencePoint(name, l);
    	}
    }
    
    public String getUnusedReferenceId() {
    	return getSourceModel().getUnusedReferenceId();
    }
    
    public void setReferencePoint(Point p, String name, Object caller) {
    	getSourceModel().setReferencePoint(p, name, caller);
    }
    
//	SourceCodeEditor
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	FocusListener
	public void focusGained(FocusEvent e) {
		m_caret.setVisible(true);
		if( getSelection() != null ) {
			getSelection().asComponent().setVisible(true);
		}
		m_bHasFocus = true;
	}

	public void focusLost(FocusEvent e) {
		m_caret.setVisible(false);
		if( getSelection() != null ) {
			getSelection().asComponent().setVisible(false);
		}
		m_bHasFocus = false;
	}
//	FocusListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	MouseListener
	public void mouseClicked(MouseEvent e) {
		if( !e.isConsumed() ) {
			if( (e.getModifiers()&InputEvent.BUTTON1_MASK) == InputEvent.BUTTON1_MASK ) {
				Point mouseCoords = e.getPoint();
				mouseCoords = SwingUtilities.convertPoint(e.getComponent(), mouseCoords, this);
 				mouseCoords.x += getFontWidth()/2;
				if( getSelection() != null && getSelection().contains(mouseCoords.x-getSelection().getLocation().x, mouseCoords.y-getSelection().getLocation().y) ) {
					UndoChain undoChain = new UndoChain("mouseClicked");
					undoChain.addRedo(clearSelection());
					Point pos = viewToPosition( mouseCoords );
					undoChain.addRedo(setCaretPosition(pos, true));
					e.consume();
					getUndoManager().addUndo(undoChain);
				}
			}
		}
	}
	
	public void mousePressed(MouseEvent e) {
//		System.err.println(e);
		if( !e.isConsumed() ) {
			if( (e.getModifiers()&InputEvent.BUTTON1_MASK) == InputEvent.BUTTON1_MASK ) {
				Point mouseCoords = e.getPoint();
				mouseCoords = SwingUtilities.convertPoint(e.getComponent(), mouseCoords, this);
				m_bSelectWordMode = false;
				m_bDragSelectionMode = false;
				requestFocus();
 				mouseCoords.x += getFontWidth()/2;
 				UndoChain undoCmd = new UndoChain("selection");
				if( e.getClickCount() == 1 ) {
					if( getSelection() != null && getSelection().contains(mouseCoords.x-getSelection().getLocation().x, mouseCoords.y-getSelection().getLocation().y) ) {
		//				startSelectionDrag(mouseCoords);
					} else {
						Point pos = viewToPosition( mouseCoords );
						if( e.isShiftDown() ) {
							undoCmd.addRedo(setSelectionMode(SourceCodeSelection.BOUNDED));
							undoCmd.addRedo(setSelectionPosition(pos));
							e.consume();
						} else if( e.isAltDown() ) {
							undoCmd.addRedo(setSelectionMode(SourceCodeSelection.BLOCKED));
							undoCmd.addRedo(setSelectionPosition(pos));
							e.consume();
						} else if( e.isControlDown() ) {
							undoCmd.addRedo(startWordSelection(mouseCoords));
							e.consume();
						} else {
							undoCmd.addRedo(setCaretPosition(pos, true));
							e.consume();
						}
					}
				} else if( e.getClickCount() == 2 ) {
						if( e.isAltDown() ) {
						} else {
							undoCmd.addRedo(startWordSelection(mouseCoords));
							e.consume();
						}
				}
				getUndoManager().addUndo(undoCmd);
			}
		}
	}
	public void mouseReleased(MouseEvent e) {
		if( !e.isConsumed() ) {
			if( (e.getModifiers()&InputEvent.BUTTON1_MASK) == InputEvent.BUTTON1_MASK ) {
				m_autoscrollController.stopVerticalAutoscroll();
				m_autoscrollController.stopHorizontalAutoscroll();
				Point mouseCoords = e.getPoint();
				mouseCoords = SwingUtilities.convertPoint(e.getComponent(), mouseCoords, this);
				if( m_bDragSelectionMode ) {
					if( m_bSelectWordMode ) {
						getUndoManager().addUndo(stopWordSelection(mouseCoords));
					} else if( m_bDragSelectionMode ) {
		//				stopSelectionDrag(mouseCoords, true);
					}
				}
			}
		}
	}
	public void mouseEntered(MouseEvent e) {
	}
	public void mouseExited(MouseEvent e) {
	}
//	MouseListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	MouseMotionListener
	public UndoCommand extendViewSelection( Point view, int modifiers ) {
		UndoChain rc = new UndoChain("selection");
		Point pos = viewToPosition( view );
		if( m_bSelectWordMode ) {
			rc.addRedo(updateWordSelection(view));
		} else {
			if( (modifiers&InputEvent.ALT_MASK) != 0 ) {
				rc.addRedo(setSelectionMode(SourceCodeSelection.BLOCKED));
				rc.addRedo(setSelectionPosition(pos));
			} else {
				rc.addRedo(setSelectionMode(SourceCodeSelection.BOUNDED));
				rc.addRedo(setSelectionPosition(pos));
			}
		}
		return rc;
	}
	
	public void mouseDragged(MouseEvent e) {
		if( !e.isConsumed() ) {
			Point mouseCoords = e.getPoint();
			mouseCoords = SwingUtilities.convertPoint(e.getComponent(), mouseCoords, this);
			Rectangle b = getEditBounds();

			// Limit selection and do autoscroll
			if( mouseCoords.x < b.x ) {
				m_autoscrollController.startHorizontalAutoscroll(mouseCoords, e.getModifiers());
				mouseCoords.x = b.x;
			} else if( mouseCoords.x > b.x + (getLastVisibleColumn()-getFirstVisibleColumn())*getFontWidth() ) {
				m_autoscrollController.startHorizontalAutoscroll(mouseCoords, e.getModifiers());
				mouseCoords.x = b.x + (getLastVisibleColumn()-getFirstVisibleColumn())*getFontWidth();
            } else {
            	m_autoscrollController.stopHorizontalAutoscroll();
            }
			if( mouseCoords.y < b.y ) {
				m_autoscrollController.startVerticalAutoscroll(mouseCoords, e.getModifiers());
				mouseCoords.y = b.y;
			} else if( mouseCoords.y > (b.y + (getLastVisibleRow()-getFirstVisibleRow())*getFontHeight()) && 
					   getLineCount() > getLastVisibleRow()) {
				m_autoscrollController.startVerticalAutoscroll(mouseCoords, e.getModifiers());
				mouseCoords.y = b.y + (getLastVisibleRow()-getFirstVisibleRow())*getFontHeight();
            } else {
            	m_autoscrollController.stopVerticalAutoscroll();
            }
            
			if( !m_bDragSelectionMode ) {
				mouseCoords.x += getFontWidth()/2;
            	getUndoManager().addUndo(extendViewSelection( mouseCoords, e.getModifiers()));
				e.consume();
			} else {
	//			updateSelectionDrag(mouseCoords);
				System.err.println(e);
			}
		}
	}

	public void mouseMoved(MouseEvent e) {
//		System.err.println(e);
	}
//	MouseMotionListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	KeyListener
	public void keyTyped(KeyEvent e) {
	}

	private static final int MODMASK = InputEvent.CTRL_MASK|InputEvent.ALT_MASK|InputEvent.SHIFT_MASK;
	public void keyPressed(KeyEvent e) {
		if( ( !e.isConsumed() ) && ( e.getModifiers()&(~InputEvent.SHIFT_MASK)&MODMASK )==0 ) {
			switch( e.getKeyCode() ) {
				case KeyEvent.VK_SPACE:
				case KeyEvent.VK_COMMA:
				case KeyEvent.VK_MINUS:
				case KeyEvent.VK_PERIOD:
				case KeyEvent.VK_SLASH:
				case KeyEvent.VK_0:
				case KeyEvent.VK_1:
				case KeyEvent.VK_2:
				case KeyEvent.VK_3:
				case KeyEvent.VK_4:
				case KeyEvent.VK_5:
				case KeyEvent.VK_6:
				case KeyEvent.VK_7:
				case KeyEvent.VK_8:
				case KeyEvent.VK_9:
				case KeyEvent.VK_SEMICOLON:
				case KeyEvent.VK_EQUALS:
				case KeyEvent.VK_A:
				case KeyEvent.VK_B:
				case KeyEvent.VK_C:
				case KeyEvent.VK_D:
				case KeyEvent.VK_E:
				case KeyEvent.VK_F:
				case KeyEvent.VK_G:
				case KeyEvent.VK_H:
				case KeyEvent.VK_I:
				case KeyEvent.VK_J:
				case KeyEvent.VK_K:
				case KeyEvent.VK_L:
				case KeyEvent.VK_M:
				case KeyEvent.VK_N:
				case KeyEvent.VK_O:
				case KeyEvent.VK_P:
				case KeyEvent.VK_Q:
				case KeyEvent.VK_R:
				case KeyEvent.VK_S:
				case KeyEvent.VK_T:
				case KeyEvent.VK_U:
				case KeyEvent.VK_V:
				case KeyEvent.VK_W:
				case KeyEvent.VK_X:
				case KeyEvent.VK_Y:
				case KeyEvent.VK_Z:
				case KeyEvent.VK_OPEN_BRACKET:
				case KeyEvent.VK_BACK_SLASH:
				case KeyEvent.VK_CLOSE_BRACKET:
				case KeyEvent.VK_MULTIPLY:
				case KeyEvent.VK_ADD:
				case KeyEvent.VK_SEPARATER:
				case KeyEvent.VK_SUBTRACT:
				case KeyEvent.VK_DECIMAL:
				case KeyEvent.VK_DIVIDE:
				case KeyEvent.VK_BACK_QUOTE:
				case KeyEvent.VK_QUOTE:
				case KeyEvent.VK_DEAD_GRAVE:
				case KeyEvent.VK_DEAD_ACUTE:
				case KeyEvent.VK_DEAD_CIRCUMFLEX:
				case KeyEvent.VK_DEAD_TILDE:
				case KeyEvent.VK_DEAD_MACRON:
				case KeyEvent.VK_DEAD_BREVE:
				case KeyEvent.VK_DEAD_ABOVEDOT:
				case KeyEvent.VK_DEAD_DIAERESIS:
				case KeyEvent.VK_DEAD_ABOVERING:
				case KeyEvent.VK_DEAD_DOUBLEACUTE:
				case KeyEvent.VK_DEAD_CARON:
				case KeyEvent.VK_DEAD_CEDILLA:
				case KeyEvent.VK_DEAD_OGONEK:
				case KeyEvent.VK_DEAD_IOTA:
				case KeyEvent.VK_DEAD_VOICED_SOUND:
				case KeyEvent.VK_DEAD_SEMIVOICED_SOUND:
				case KeyEvent.VK_AMPERSAND:
				case KeyEvent.VK_ASTERISK:
				case KeyEvent.VK_QUOTEDBL:
				case KeyEvent.VK_LESS:
				case KeyEvent.VK_GREATER:
				case KeyEvent.VK_BRACELEFT:
				case KeyEvent.VK_BRACERIGHT:
				case KeyEvent.VK_AT:
				case KeyEvent.VK_COLON:
				case KeyEvent.VK_CIRCUMFLEX:
				case KeyEvent.VK_DOLLAR:
				case KeyEvent.VK_EURO_SIGN:
				case KeyEvent.VK_EXCLAMATION_MARK:
				case KeyEvent.VK_INVERTED_EXCLAMATION_MARK:
				case KeyEvent.VK_LEFT_PARENTHESIS:
				case KeyEvent.VK_NUMBER_SIGN:
				case KeyEvent.VK_PLUS:
				case KeyEvent.VK_RIGHT_PARENTHESIS:
				case KeyEvent.VK_UNDERSCORE: {
					UndoCommand undoCmd = null;
						
					if( isInsert() ) {
						undoCmd = insertChar(e.getKeyChar());
					} else if( isOverwrite() ) {
						undoCmd = overwriteChar(e.getKeyChar());
					}
						
					e.consume();
					if( undoCmd != null ) {
						m_sourceModel.getUndoManager().addUndo(undoCmd);
					}
				} break;
			}
		}
	}

	public void keyReleased(KeyEvent e) {
	}
//	KeyListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	ChangeListener
	public void stateChanged(ChangeEvent e) {
		m_caret.updateBounds();
	}
//	ChangeListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	LayoutManager
	public void addLayoutComponent(String name, Component comp) {
		return;
	}

	public void layoutContainer(Container parent) {
		if( parent == this ) {
			m_bar.updateBounds();
			if( m_selection != null ) {
				m_selection.updateBounds();
			}
			m_caret.updateBounds();
		}
	}

	public Dimension minimumLayoutSize(Container parent) {
		Dimension rc = new Dimension(0,0);
		return rc;
	}

	public Dimension preferredLayoutSize(Container parent) {
		Dimension rc = new Dimension(0,0);
		return rc;
	}

	public void removeLayoutComponent(Component comp) {
	}
//	LayoutManager
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	DropTargetListener
	public void dragEnter(DropTargetDragEvent dtde) {
	    startSelectionDrag(dtde);
	    repaint();
	}
	public void dragOver(DropTargetDragEvent dtde) {
	    updateSelectionDrag(dtde);
	}
	public void dropActionChanged(DropTargetDragEvent dtde) {
		int action = dtde.getDropAction();
    	if( (action&DnDConstants.ACTION_COPY_OR_MOVE) != 0 ) {
			dtde.acceptDrag(dtde.getDropAction());
    	} else {
    		dtde.rejectDrag();
    	}
	}
	public void dragExit(DropTargetEvent dte) {
	    cancelSelectionDrag();
	}
	public void drop(DropTargetDropEvent dtde) {
	    completeSelectionDrag(dtde);
	    repaint();
	}
//	DropTargetListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	Autoscroll
	public Insets getAutoscrollInsets() {
		return new Insets(getFontHeight()/2, m_bar.getWidth()+getFontWidth()/2, getFontHeight()/2, getFontWidth()/2);
	}
	
	public void autoscroll(Point cursorLocn) {
     	synchronized(this) {
    		m_bDelayCancelSelection = false;
    		notify();
    	}
    	Insets insets = getAutoscrollInsets();
    	if( cursorLocn.y < insets.top ) {
    		getEditCommands().windowScrollDown();
    	} else if ( cursorLocn.y > (getHeight()-insets.bottom ) ){
    		getEditCommands().windowScrollUp();
    	}
    	if( cursorLocn.x < insets.left ) {
    		getEditCommands().windowScrollLeft();
    	} else if ( cursorLocn.x > (getWidth()-insets.right ) ){
    		getEditCommands().windowScrollRight();
    	}
    	Point view = new Point(cursorLocn);
    	view.x += getFontWidth()/2;
    	Point pos = viewToPosition(view);
 		UndoCommand undoCmd = null;
    	if( ( getSelection() != null ) && 
    		( getSelection().asComponent().isVisible()) && 
    		( getSelection().containsPos(pos) ) ) {
    		m_caret.setVisible(false);
    	} else {
    		m_caret.setVisible(true);
    		undoCmd = setCaretPosition(pos, false);
    	}
		getUndoManager().addUndo(undoCmd);
	}
//	Autoscroll
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//	ContainerListener
	public void componentAdded(ContainerEvent e) {
		e.getChild().addMouseListener(this);
	}
	
	public void componentRemoved(ContainerEvent e) {
		e.getChild().removeMouseListener(this);
	}
//	ContainerListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  DirtListener
	public void dirtyChanged( boolean bDirty, DirtManager mgr ) {
		notifyDirty();
	}
	
	protected void notifyDirty() {
    	Iterator iter = m_dirtListeners.iterator();
    	while( iter.hasNext() ) {
    		DirtListener l = (DirtListener)iter.next();
    		l.dirtyChanged( isDirty(), this );
    	}
	}
//  DirtListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  DirtManager
	public boolean isDirty() {
		return getSourceModel().isDirty();
	}
	
	public void clearDirty() {
		getSourceModel().isDirty();
	}
	
    public void addDirtListener( DirtListener l ) {    	
    	if( !m_dirtListeners.contains(l) ) {
    		m_dirtListeners.add(l);
    	} else {
    		System.err.println("View already contains listener.");
    	}
    }
    
    public void removeDirtListener( DirtListener l ) {
    	m_dirtListeners.remove(l);
    }
//  DirtManager
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  StatusSource
	private void setStatusLocation( Point pos ) {
		pos = posToDoc(docToPos(pos));
		StringBuffer str = new StringBuffer("Ln "+(pos.y+1)+", Col "+(pos.x+1));
		while( str.length() < 14 ) {
			str.append(" ");
		}
		setStatusString(0, str.toString());
	}
	
	public void setRecordStatus( boolean active ) {
		setStatusActive(1, active);
	}
	
	public void setColumnStatus( boolean active ) {
		setStatusActive(2, active);
	}
	
	private void setOverwriteStatus( boolean active ) {
		setStatusActive(3, active);
	}
	
	public void setReadonlyStatus( boolean active ) {
		setStatusActive(4, active);
	}
	
	public void addStatusListener( StatusListener l ) {
		m_statusListeners.addElement(l);
	}
	
	public void removeStatusListener( StatusListener l ) {
		m_statusListeners.removeElement(l);
	}
	
	public String getStatusMsg() {
		return m_statusMsg;
	}
	
	public void setStatusMsg(String str ) {
		if( !m_statusMsg.equals(str) ) {
			m_statusMsg = str;
			Enumeration enum = m_statusListeners.elements();
			while( enum.hasMoreElements() ) {
				StatusListener l = (StatusListener)enum.nextElement();
				l.statusMessageChanged();
			}
		}
	}
			
	public int getStatusCount() {
		return 5;
	}
	
	public String getStatusString(int index) {
		return m_statusStr[index];
	}
	
	public boolean isStatusActive(int index) {
		return m_statusActive[index];
	}
	
	private void setStatusString( int index, String str ) {
		if( !m_statusStr[index].equals(str) ) {
			m_statusStr[index] = str;
			notifyStatusListeners(index, m_statusStr[index], m_statusActive[index]);
		}		
	}
	
	private void setStatusActive( int index, boolean active ) {
		if( m_statusActive[index] != active ) {
			m_statusActive[index] = active;
			notifyStatusListeners(index, m_statusStr[index], m_statusActive[index]);
		}
	}
	
	private void notifyStatusListeners(int index, String str, boolean active ) {
		Enumeration enum = m_statusListeners.elements();
		while( enum.hasMoreElements() ) {
			StatusListener l = (StatusListener)enum.nextElement();
			l.statusChanged(index, str, active);
		}
	}		
//  StatusSource
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//  EditCaretListener
    public void caretUpdate( EditCaretEvent e ) {
    	if( e.isPositionChange() ) {
    		Point p = e.getPosition();
    		setStatusLocation(p);
			scrollToCaret();    		
    	} else if( e.isModeChange() ) {
    		if( e.getMode() == EditCaret.INSERTMODE ) {
    			setOverwriteStatus( false );
    		} else if( e.getMode() == EditCaret.OVERWRITEMODE ) {
    			setOverwriteStatus( true );
    		}
    	}
    }
//  EditCaretListener
///////////////////////////////////////////////////////////////////////////////

	public static synchronized void buildCursor(Component c) {
		if( CursorManager.getInstance().getCursor("DRAGMOVE") == null ||
			CursorManager.getInstance().getCursor("DRAGCOPY") == null ) {
            MediaTracker tracker = new MediaTracker(c);
            Image dragMove = Toolkit.getDefaultToolkit().getImage(ClassLoader.getSystemResource("gvid/ui/icons/DRAGMOVE.gif"));
            Image dragCopy = Toolkit.getDefaultToolkit().getImage(ClassLoader.getSystemResource("gvid/ui/icons/DRAGCOPY.gif"));
            tracker.addImage(dragMove, 0);
            tracker.addImage(dragCopy, 1);
            try {
                tracker.waitForID(0);
                tracker.waitForID(1);
            } catch (InterruptedException ie) {
                ie.printStackTrace();
            }
            Toolkit.getDefaultToolkit().createCustomCursor(dragMove, new Point(16,8), "DRAGMOVE");
            Toolkit.getDefaultToolkit().createCustomCursor(dragCopy, new Point(16,8), "DRAGCOPY");
		}
	}
	
	
	// returns true if p1 is before p2
	public boolean isBefore( Point p1, Point p2 ) {
		boolean rc = false;
		if( p1.y < p2.y ) {
			rc = true;
		} else if( p1.y==p2.y && p1.x < p2.x ) {
			rc = true;
		}		
		return rc;
	}
	
	public static void main( String[] args ) {
		String filename = "gvid/ui/editor/SourceCodeView.java";

		if( args.length > 0 ) {
			filename = args[0];
		}

		try {
//			  UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
//			  com.sun.java.swing.plaf.windows.WindowsLookAndFeel laf = new com.sun.java.swing.plaf.windows.WindowsLookAndFeel();
			LookAndFeel laf = new com.sun.java.swing.plaf.windows.WindowsLookAndFeel() {
				public boolean isNativeLookAndFeel() {
					return true;
				}
			};
			UIManager.setLookAndFeel(laf);
		} catch( Exception e ) {
			System.err.println(e);
		}

//		final gvid.ui.editor.SourceCodeView scv = test.gvid.ui.editor.SourceCodeView.buildSourceCodeView();
		final SourceCodeView scv = new SourceCodeView();
		scv.loadFile(filename);
//		scv.setCaretPosition(3,0);
//		scv.setSelectionPosition(new Point(5,8), SourceCodeSelection.BOUNDED);
		
		final SplitScrollPane sp = new SplitScrollPane();
		sp.setView(scv);
		
		final JFrame f = new JFrame();
		f.setTitle("ViD - "+filename);
		f.getContentPane().setLayout(new BorderLayout());
		f.addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				f.dispose();
			}

			public void windowClosed(WindowEvent e) {
				System.exit(0);
			}
		});
		f.getRootPane().addFocusListener( new FocusListener() {
			public void focusGained(FocusEvent e) {
				sp.requestFocus();
			}
			
			public void focusLost(FocusEvent e) {
			}
		});
		
		final String titleName = filename;
		scv.addDirtListener( new DirtListener() {
			public void dirtyChanged( boolean bDirty, DirtManager mgr ) {
				if( bDirty ) {
					f.setTitle("ViD - "+titleName+"*");
				} else {
					f.setTitle("ViD - "+titleName);
				}
			}
		});
		
		f.getContentPane().add( sp, BorderLayout.CENTER );
		
		f.setSize(500,400);
		f.doLayout();

		f.setVisible(true);
		
		return;
	}
}