/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.ui.splitscrollpane;

import javax.swing.*;
import java.beans.*;
import java.awt.*;
import java.awt.event.*;
import gvid.util.CursorManager;

/**
 * A control that allows the user to split a view into two views.
 * 
 * @author Alan Oursland
 * @version 1.00
 * @since 02/22/2001
 */
public class SplitControl 
    extends     javax.swing.JComponent
    implements  MouseListener,
                MouseMotionListener,
                KeyListener
{

    /**
     * Specifies the alignment of the split control.
     */
    private int m_alignment = SwingConstants.HORIZONTAL;
    private int m_state = INACTIVE;
    private int m_inactiveEdge = 0;
    private int m_activeEdge = 0;
    private int m_inactivePosition = 0;
    private int m_position = 0;
    private int m_maxPosition = 0;
    private boolean m_bMoving = false;
    private boolean m_bDualDrag = false;
    private Point m_lastMousePosition = new Point();
    transient ActionListener m_actionListener;
    
    private SplitMover m_mover = new SplitMover(this);
    
    public final static int DEFAULTSIZE = 7;
    
	public final static int INACTIVE = 0;
	public final static int ACTIVE = 1;
	
	public final static int POSITIONACTION = 0;
	public final static int STATEACTION = 1;
	
	/**
	 * @param alignment SwingConstants.VERTICAL or SwingConstants.HORIZONTAL
	 */
	public SplitControl(int alignment)
	{
	    setBorder(null);
	    setOpaque(true);
	    m_alignment = alignment;
	    buildSplitCursors(this);
	    addMouseListener(this);
	    addMouseMotionListener(this);
	    
	    setState(INACTIVE);
	    if( m_alignment == SwingConstants.VERTICAL ) {
	        setPreferredSize( new Dimension( 16, DEFAULTSIZE ) );
	        setMinimumSize( new Dimension( 16, DEFAULTSIZE ) );
	    } else {
	        setPreferredSize( new Dimension( DEFAULTSIZE, 16 ) );
	        setMinimumSize( new Dimension( DEFAULTSIZE, 16 ) );
	    }
	}
	
	public int getAlignment() {
	    return m_alignment;
	}
	
	public void setInactiveEdge( int edge ) {
	    m_inactiveEdge = edge;
	}
	
	public void setActiveEdge( int edge ) {
	    m_activeEdge = edge;
	}
	
	public int getActiveEdge() {
	    return m_activeEdge;
	}
	
	public void setInactivePosition( int position ) {
	    m_inactivePosition = position;
	}
	
	public int getInactivePosition() {
	    return m_inactivePosition;
	}
	
	public int getPosition() {
	    return m_position;
	}
	
	public void setPosition( int position ) {
	    if( position != m_position ) {
	        m_position = position;
	        validatePosition();
/*	        
            if( m_position == m_inactivePosition ) {
                setState(INACTIVE);
            } else {
                setState(ACTIVE);
            }
*/            
            if( m_alignment == SwingConstants.HORIZONTAL ) {
                processActionEvent( new ActionEvent(this, POSITIONACTION, "HSPLIT") );
            } else {
                processActionEvent( new ActionEvent(this, POSITIONACTION, "VSPLIT") );
            }
            if( getParent() != null ) {
                getParent().validate();
            }
	    }
	}
	
	public int getMaxPosition() {
	    return m_maxPosition;
	}
	
	public void setMaxPosition( int max ) {
	    m_maxPosition = max;
	}
	
	public int getState() {
	    return m_state;
	}
	
	public boolean isActive() {
	    return (m_state==ACTIVE);
	}
	
	public void addActionListener( ActionListener l ) {
	    
	    if (l == null) {
	        return;
	    }
	    m_actionListener = AWTEventMulticaster.add(m_actionListener, l);
	    enableEvents(AWTEvent.ACTION_EVENT_MASK);
	}
	
	public void removeActionListener( ActionListener l ) {
	    if (l == null) {
	        return;
	    }
	    m_actionListener = AWTEventMulticaster.remove(m_actionListener, l);
	    if( m_actionListener == null ) {
	        disableEvents(AWTEvent.ACTION_EVENT_MASK);
	    }
	}
	
	public void updateBounds() {
//	    System.err.println("Updating split bounds");
        validatePosition();
        if( m_bMoving ) {
            m_mover.updateBounds();
        }
	    switch( m_state ) {
            case INACTIVE: {
                Rectangle b = getDragBoxBounds();
                setBounds( b );
            } break;
            case ACTIVE: {
	            Insets insets = getParent().getInsets();
	            Dimension size = getParent().getSize();
                if( m_alignment == SwingConstants.HORIZONTAL ) {
	                setBounds(insets.left+m_position, m_activeEdge-insets.top, SplitControl.DEFAULTSIZE, size.height);
                } else {
	                setBounds(m_activeEdge-insets.left, insets.top+m_position, size.width, SplitControl.DEFAULTSIZE);
                }
            } break;
	    }
	}
	
	public void paint( Graphics g ) {
	    super.paint(g);
	    
        switch( m_state ) {
            case INACTIVE: {
                drawDragBox(g);
            } break;
            case ACTIVE: {
                drawDragBar(g);
            } break;
        }
	}
	
	public void setState( int state ) {
	    if( state != m_state ) {
	        m_state = state;
            switch( m_state ) {
                case INACTIVE: {
                } break;
                case ACTIVE: {
/*                    
                    if( m_position == m_maxPosition ) {
                        m_position = m_inactivePosition;
                        setState(INACTIVE);
                        return;
                    }
*/                    
                } break;
            }
            if( m_alignment == SwingConstants.HORIZONTAL ) {
                processActionEvent( new ActionEvent(this, STATEACTION, "HSPLIT") );
            } else {
                processActionEvent( new ActionEvent(this, STATEACTION, "VSPLIT") );
            }
            if( getParent() != null ) {
                getParent().validate();
            }
	    }
	}
	
    protected void processEvent(AWTEvent e) {
        if (e instanceof ActionEvent) {
            processActionEvent((ActionEvent)e);
            return;
        }
	    super.processEvent(e);
    }

    protected void processActionEvent(ActionEvent e) {
//        dispatchEvent(e);
        if (m_actionListener != null) {
            m_actionListener.actionPerformed(e);
        }
    }
    
    private void validatePosition() {
        if( m_position < m_inactivePosition ) {
            m_position = m_inactivePosition;
        }
        if( m_position > m_maxPosition ) {
            m_position = m_maxPosition;
        }
    }
	private void drawDragBox( Graphics g ) {
	    Rectangle b = getBounds();
	    Rectangle d = getDragBoxBounds();
	    Rectangle r = new Rectangle( d.x-b.x, d.y-b.y, d.width, d.height );
	    
	    Color c = g.getColor();
	    
	    g.clearRect(r.x, r.y, r.width, r.height);
	    
	    g.setColor(Color.black);
	    g.drawLine(r.x,           r.y+r.height-1, r.x+r.width-1, r.y+r.height-1);
	    g.drawLine(r.x+d.width-1, r.y,            r.x+d.width-1, r.y+d.height-1);
            	    
	    g.setColor(Color.gray);
	    g.drawLine(r.x+1,         r.y+d.height-2, r.x+d.width-2, r.y+d.height-2);
	    g.drawLine(r.x+d.width-2, r.y+1,          r.x+d.width-2, r.y+d.height-2);
            	    
	    g.setColor(Color.white);
	    g.drawLine(r.x+1, r.y+1, r.x+d.width-2, r.y+1);
	    g.drawLine(r.x+1, r.y+1, r.x+1,         r.y+d.height-2);
	    
	    g.setColor(c);
	}
	
	private void drawDragBar( Graphics g ) {
	    Rectangle b = getBounds();
	    
	    Color c = g.getColor();
	    
	    g.clearRect(0, 0, b.width, b.height);
	    
	    // There may be an intersecting split control. Check the parent's layout
	    Rectangle isect = getOrthogIntersection();
	    
	    g.setColor(Color.black);
	    if( m_alignment == SwingConstants.VERTICAL ) {
	        // draw bar border
	        g.drawLine(1,                   b.height-1, isect.x-1, b.height-1);
	        g.drawLine(isect.x+isect.width-1, b.height-1, b.width-3, b.height-1);
	    } else {
	        // draw bar border
	        g.drawLine(b.width-1, 1,                      b.width-1, isect.y-1);
	        g.drawLine(b.width-1, isect.y+isect.height-1, b.width-1, b.height-3);
	    }
            	    
	    g.setColor(Color.gray);
	    if( m_alignment == SwingConstants.VERTICAL ) {
	        // draw bar border
	        g.drawLine(0,                     b.height-2, isect.x-1, b.height-2);
	        g.drawLine(isect.x+isect.width-1, b.height-2, b.width-2, b.height-2);
	        
	        // connect to main border
	        g.drawLine(0,         b.height-2, 0,         b.height);
	        g.drawLine(0,         0,          0,         1);
	    } else {
	        // draw bar border
	        g.drawLine(b.width-2, 0,                      b.width-2, isect.y-1);
	        g.drawLine(b.width-2, isect.y+isect.height-1, b.width-2, b.height-2);
	        
	        // connect to main border
	        g.drawLine(b.width-2, 0,          b.width, 0);
	        g.drawLine(0,         0,          1,       0);
	    }
	    // if intersection, connect intersection borders
	    if( isect.width > 0 || isect.height > 0 ) {
	        g.drawLine(isect.x+isect.width-2, isect.y,                isect.x+isect.width-2, isect.y);
	        g.drawLine(isect.x+isect.width-2, isect.y+isect.height-2, isect.x+isect.width-2, isect.y+isect.height-1);
	        g.drawLine(isect.x+isect.width-2, isect.y+isect.height-2, isect.x+isect.width-1, isect.y+isect.height-2);
	        g.drawLine(isect.x,               isect.y+isect.height-2, isect.x,               isect.y+isect.height-2);
	    }
            	    
	    g.setColor(Color.white);
	    if( m_alignment == SwingConstants.VERTICAL ) {
	        // draw bar border
	        g.drawLine(1, 1, isect.x+1, 1);
	        g.drawLine(isect.x+isect.width-2, 1, b.width-2, 1);
	        
	        // connect to main border
	        g.drawLine(b.width-1, b.height-2, b.width-1, b.height);
	        g.drawLine(b.width-1, 0,          b.width-1, 1);
	    } else {
	        // draw bar border
	        g.drawLine(1, 0,                    1, isect.y+1);
	        g.drawLine(1, isect.y+isect.height-2, 1, b.height-2);
	        
	        // connect to main border
	        g.drawLine(b.width-2, b.height-1, b.width, b.height-1);
	        g.drawLine(0,         b.height-1, 1,       b.height-1);
	    }
	    if( isect.width > 0 || isect.height > 0 ) {
            g.drawLine(isect.x, isect.y+1, isect.x, isect.y+1);
            g.drawLine(isect.x+1, isect.y, isect.x+1, isect.y+1);
            g.drawLine(isect.x+isect.width-2, isect.y+1, isect.x+isect.width-1, isect.y+1);
            g.drawLine(isect.x+1, isect.y+isect.height-2, isect.x+1, isect.y+isect.height-1);
        }	    
	    g.setColor(c);
	}
	
	private SplitControl getOrthogSplit() {
	    SplitControl rc = null;
	    if( getParent() != null ) {
	        LayoutManager lm = getParent().getLayout();
	        if( lm instanceof SplitScrollPaneLayout ) {
	            SplitScrollPaneLayout sspl = (SplitScrollPaneLayout)lm;
	            if( m_alignment == SwingConstants.VERTICAL ) {
	                rc = (SplitControl)sspl.getLayoutComponent(SplitScrollPaneLayout.HSPLIT);
	            } else {
	                rc = (SplitControl)sspl.getLayoutComponent(SplitScrollPaneLayout.VSPLIT);
	            }
	        }
	    }
	    return rc;
	}
	
	private Rectangle getOrthogIntersection() {
	    Rectangle rc = new Rectangle(0,0,0,0);
	    if( isActive() ) {
	        SplitControl sc = getOrthogSplit();
	        if( sc != null && sc.isActive() ) {
	            rc = getBounds();
	            rc = rc.intersection(sc.getBounds());
	            rc.x -= getBounds().x;
	            rc.y -= getBounds().y;
	        }
	    }
	    return rc;
	}
	
	private Rectangle getDragBoxBounds() {
	    Rectangle rc = null;
	    if( getParent() != null ) {
	        Insets insets = getParent().getInsets();
	        Dimension size = getParent().getSize();
	        if( m_alignment == SwingConstants.HORIZONTAL ) {
	            rc = new Rectangle(insets.left+m_inactivePosition, m_inactiveEdge, SplitControl.DEFAULTSIZE, size.height-m_inactiveEdge-insets.bottom);
	        } else {
	            rc = new Rectangle(m_inactiveEdge, insets.top+m_inactivePosition, size.width-m_inactiveEdge-insets.right, SplitControl.DEFAULTSIZE);
	        }
	    } else {
	        rc = new Rectangle();
	    }
	    return rc;
	}
	
	private void setAppropriateCursor( Point mousePoint ) {
        if( m_bMoving ) {
            if( m_bDualDrag ) {
                m_mover.setCursor(CursorManager.getInstance().getCursor("HV_SPLIT"));
                setCursor(CursorManager.getInstance().getCursor("HV_SPLIT"));
            } else if( m_alignment == SwingConstants.HORIZONTAL ) {
                m_mover.setCursor(CursorManager.getInstance().getCursor("HO_SPLIT"));
                setCursor(CursorManager.getInstance().getCursor("HO_SPLIT"));
            } else {
                m_mover.setCursor(CursorManager.getInstance().getCursor("VE_SPLIT"));
                setCursor(CursorManager.getInstance().getCursor("VE_SPLIT"));
            }
	    } else if( contains(mousePoint) ) {
	        Rectangle isect = getOrthogIntersection();
            if( isect.contains(mousePoint) ) {
                setCursor(CursorManager.getInstance().getCursor("HV_SPLIT"));
            } else {
                switch( m_alignment ) {
                    case SwingConstants.HORIZONTAL: {
                        setCursor(CursorManager.getInstance().getCursor("HO_SPLIT"));
                    } break;
                    case SwingConstants.VERTICAL: {
                        setCursor(CursorManager.getInstance().getCursor("VE_SPLIT"));
                    }
                }
            }
	    } else {
	        setCursor(Cursor.getDefaultCursor());
	    }
	}
	
	private void startMoving() {
	    if( getParent() != null ) {
	        m_bMoving = true;
	        if( m_state == INACTIVE ) {
	            m_mover.setPosition(m_inactivePosition);
	        } else {
	            m_mover.setPosition(m_position);
	        }
	        getParent().add(m_mover, 0);
	        m_mover.setVisible(true);
	        m_mover.updateBounds();
	        m_mover.requestFocus();
	        m_mover.addKeyListener(this);
	    }
	}
	
	private void stopMoving() {
	    m_bMoving = false;
	    if( m_bDualDrag ) {
            SplitControl orthog = getOrthogSplit();
            if( orthog != null ) {
                m_bDualDrag = true;
                orthog.stopMoving();
            }
            m_bDualDrag = false;
	    }
	    if( getParent() != null ) {
	        getParent().remove(m_mover);
	        Rectangle b = m_mover.getBounds();
	        getParent().repaint(b.x, b.y, b.width, b.height);
	        m_mover.removeKeyListener(this);
	        getParent().validate();
	    }
	}
	
	private static synchronized void buildSplitCursors(Component c) {
		if( CursorManager.getInstance().getCursor("HO_SPLIT") == null ||
			CursorManager.getInstance().getCursor("VE_SPLIT") == null ||
			CursorManager.getInstance().getCursor("HV_SPLIT") == null ) {
            MediaTracker tracker = new MediaTracker(c);
            Image hSplitImage = Toolkit.getDefaultToolkit().getImage( ClassLoader.getSystemResource("gvid/ui/icons/VE_SPLIT.gif"));
            Image vSplitImage = Toolkit.getDefaultToolkit().getImage( ClassLoader.getSystemResource("gvid/ui/icons/HO_SPLIT.gif"));
            Image hvSplitImage = Toolkit.getDefaultToolkit().getImage(ClassLoader.getSystemResource("gvid/ui/icons/HV_SPLIT.gif"));
            tracker.addImage(hSplitImage, 0);
            tracker.addImage(vSplitImage, 1);
            tracker.addImage(hvSplitImage, 2);
            try {
                tracker.waitForID(0);
                tracker.waitForID(1);
                tracker.waitForID(2);
            } catch (InterruptedException ie) {
                ie.printStackTrace();
            }
            Toolkit.getDefaultToolkit().createCustomCursor(hSplitImage, new Point(15,15), "HO_SPLIT");
            Toolkit.getDefaultToolkit().createCustomCursor(vSplitImage, new Point(15,15), "VE_SPLIT");
            Toolkit.getDefaultToolkit().createCustomCursor(hvSplitImage, new Point(9,9), "HV_SPLIT");
		}
	}
	
	
///////////////////////////////////////////////////////////////////////////////
// MouseListener
    public void mouseClicked(MouseEvent e) {
        m_lastMousePosition = e.getPoint();        
    }

    public void mousePressed(MouseEvent e) {
        m_bDualDrag = false;
        Rectangle isect = getOrthogIntersection();
        if( isect.contains(e.getPoint()) ) {
            // handle the case where both move at the same time
            SplitControl orthog = getOrthogSplit();
            if( orthog != null ) {
                m_bDualDrag = true;
                orthog.startMoving();
            }
        }
        startMoving();
        m_lastMousePosition = e.getPoint();
        e.consume();
    }

    public void mouseReleased(MouseEvent e) {
        if( m_bMoving ) {
            if( m_bDualDrag ) {
                // handle the case where both move at the same time
                SplitControl orthog = getOrthogSplit();
                if( orthog != null ) {
//                    orthog.stopMoving();
                    orthog.setPosition(orthog.m_mover.getPosition());
                }
            }
            Point p = e.getPoint();
            // translate mouse point to parent coordinates
            p.x += getBounds().x;   
            p.y += getBounds().y;
            // move the split control
            setPosition(m_mover.getPosition());
            // translate mouse point to new local coordinates
            p.x -= getBounds().x;
            p.y -= getBounds().y;
            setAppropriateCursor(p);
            m_lastMousePosition = e.getPoint();
            e.consume();
            
            stopMoving();
        }
    }
    
    public void mouseEntered(MouseEvent e) {
        setAppropriateCursor(e.getPoint());
        m_lastMousePosition = e.getPoint();
        e.consume();
    }
    
    public void mouseExited(MouseEvent e) {
        setAppropriateCursor(e.getPoint());
        m_lastMousePosition = e.getPoint();
        e.consume();
    }
    
// MouseListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// MouseMotionListener
    public void mouseDragged(MouseEvent e) {
        if( m_bDualDrag ) {
            // handle the case where both move at the same time
            SplitControl orthog = getOrthogSplit();
            if( orthog != null ) {
                Point p = e.getPoint();
                p.x += getBounds().x - orthog.m_mover.getBounds().x;
                p.y += getBounds().y - orthog.m_mover.getBounds().y;
                if( orthog.m_alignment == SwingConstants.VERTICAL ) {
                    orthog.m_mover.move(p.y-2);
                } else {
                    orthog.m_mover.move(p.x-1);
                }
            }
        }
        
        {
            Point p = e.getPoint();
            p.x += getBounds().x - m_mover.getBounds().x;
            p.y += getBounds().y - m_mover.getBounds().y;
            if( m_alignment == SwingConstants.VERTICAL ) {
                m_mover.move(p.y-2);
            } else {
                m_mover.move(p.x-1);
            }
        }

        setAppropriateCursor(e.getPoint());
        m_lastMousePosition = e.getPoint();
        e.consume();
    }
    
    public void mouseMoved(MouseEvent e) {
        setAppropriateCursor(e.getPoint());
        m_lastMousePosition = e.getPoint();
    }
// MouseMotionListener
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// KeyListener
    public void keyPressed(KeyEvent e) {
    }
    
    public void keyReleased(KeyEvent e) {
    }

    public void keyTyped(KeyEvent e) {
        if( e.getKeyChar() == KeyEvent.VK_ESCAPE ) {
            stopMoving();
            setAppropriateCursor(m_lastMousePosition);
            e.consume();
        }
    }
// KeyListener
///////////////////////////////////////////////////////////////////////////////
}