/******************************************************************************
 * GViD Source Code Editor
 * Copyright (C) 2002  Alan Oursland
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Alan Oursland
 * alan@oursland.net
 *****************************************************************************/
package gvid.util;

import java.awt.Point;
import java.util.*;

public class ReferencePointManager {		
	private Hashtable m_refTable = new Hashtable();
	private int m_lockCount = 0;
	private Thread m_owner = null;
	
	public boolean addReference( Point p, String name, ReferencePointListener l ) {
		boolean rc = false;		
		if( !m_refTable.containsKey(name) ) {
			m_refTable.put(name, new ReferencePoint(p, l) );
			rc = true;
		}
		return rc;
	}
	
	public void removeReference( String name, ReferencePointListener l ) {
		m_refTable.remove(name);
	}
	
	public void moveReferencePointFrom( String name, ReferencePointManager man ) {
		ReferencePoint rp = (ReferencePoint)man.m_refTable.remove(name);
		if( rp != null ) {
			while( m_refTable.containsKey(name) ) {
				name = rp.m_listener.changeReferenceName(name);
			}
			m_refTable.put(name, rp);
		}
	}
	
	public void addLock() {
		synchronized(this) {
			if( ( m_owner != null ) && ( m_owner != Thread.currentThread() ) ) {
				try {
					wait();
				} catch( java.lang.InterruptedException e ) {
					System.err.println(e);
				}				
			}
			m_lockCount++;
			m_owner = Thread.currentThread();
		}
	}
	
	public void removeLock() {
		synchronized(this) {
			m_lockCount--;
			if( m_lockCount == 0 ) {
				m_owner = null;
				notify();
				fireChanges();
			}
		}
	}
	
	public Point getPoint( String name ) {
		Point rc = null;
		ReferencePoint rp = (ReferencePoint)m_refTable.get(name);
		if( rp != null ) {
			rc = new Point(rp.m_p);
		}
		return rc;
	}
	
	public void setPoint( Point p, String name, Object caller ) {
		ReferencePoint rp = (ReferencePoint)m_refTable.get(name);
		if( rp!=null && !p.equals(rp.m_p) ) {
			rp.m_p.x = p.x;
			rp.m_p.y = p.y;
			
			// If the class listening to this point
			// is making the change, assume that
			// it does not need to be notified
			if( caller == rp.m_listener ) {
				rp.clearDirty();
			}
		}
	}
	
	public boolean hasReference( String name ) {
		return m_refTable.containsKey(name);
	}
	
	public String[] getReferenceNames() {
		String[] rc = new String[m_refTable.size()];
		Enumeration enum = m_refTable.keys();
		int index = 0;
		while( enum.hasMoreElements() ) {
			String str = (String)enum.nextElement();
			rc[index++] = str;
		}
		return rc;
	}
	
	// called when splitting two lines
	public void splitRowAt(Point p) {
		Enumeration enum = m_refTable.elements();
		while( enum.hasMoreElements() ) {
			ReferencePoint rp = (ReferencePoint)enum.nextElement();
	    	if( rp.m_p.y == p.y ) {
	    		if( rp.m_p.x > p.x ) {
	    			rp.m_p.y++;
	    			rp.m_p.x -= p.x;
	    		} else if( rp.m_p.x == p.x ) {
	    			// we'll put it on the next line for kicks
	    			rp.m_p.y++;
	    			rp.m_p.x = 0;
	    		}
	    	} else if( rp.m_p.y > p.y ) {
	    		rp.m_p.y++;
	    	}
		}		
	}
	
	public void joinRowAt(int row, int rowLength) {
		Enumeration enum = m_refTable.elements();
		while( enum.hasMoreElements() ) {
			ReferencePoint rp = (ReferencePoint)enum.nextElement();
    		if( rp.m_p.y == row+1 ) {
    			rp.m_p.y--;
    			rp.m_p.x += rowLength;
    		} else if( rp.m_p.y > row+1 ) {
    			rp.m_p.y--;
    		}
		}		
	}
	
	// called when adding a character
	public void incrementColumnsAfter(Point p, int count) {
		Enumeration enum = m_refTable.elements();
		while( enum.hasMoreElements() ) {
			ReferencePoint rp = (ReferencePoint)enum.nextElement();
	    	if( rp.m_p.y == p.y && rp.m_p.x >= p.x ) {
	    		rp.m_p.x += count;
	    	}
		}		
	}
	
	public void incrementRowsAfter(int row, int count) {
		Enumeration enum = m_refTable.elements();
		while( enum.hasMoreElements() ) {
			ReferencePoint rp = (ReferencePoint)enum.nextElement();
        	if( rp.m_p.y > row ) {
        		rp.m_p.y+=count;
        	}
		}		
	}
	
	// called when deleting a character
	public void decrementColumnsAfter(Point p, int count) {
		Enumeration enum = m_refTable.elements();
		while( enum.hasMoreElements() ) {
			ReferencePoint rp = (ReferencePoint)enum.nextElement();
	    	if( rp.m_p.y == p.y && rp.m_p.x >= p.x ) {
	    		rp.m_p.x -= count;
	    		// Make sure the reference does not go past the deletion point
	    		rp.m_p.x = Math.max( rp.m_p.x, p.x );
	    	}
		}		
	}
	
	// called when deleting a character
	public void decrementRowsAfter(int row, int count) {
		Enumeration enum = m_refTable.elements();
		while( enum.hasMoreElements() ) {
			ReferencePoint rp = (ReferencePoint)enum.nextElement();
    		if( rp.m_p.y > row && rp.m_p.y <= row+count ) {
    			rp.m_p.x = 0;
    			rp.m_p.y = row+1;
    		} else if( rp.m_p.y > row+count ) {
    			rp.m_p.y -= count;
    		}
		}		
	}
	
	int m_stringIndex = 0;
	public String getUnusedString() {
		while( m_refTable.containsKey(""+m_stringIndex) ) {
			m_stringIndex++;
		}
		return ""+m_stringIndex++;
	}
	
	private void fireChanges() {
		Enumeration enum = m_refTable.keys();
		while( enum.hasMoreElements() ) {
			String key = (String)enum.nextElement();
			ReferencePoint rp = (ReferencePoint)m_refTable.get(key);
			if( rp.isDirty() == true ) {
				rp.clearDirty();
				if( rp.m_listener != null ) {
					rp.m_listener.pointChanged(new Point(rp.m_p), key);
				}
			}
		}
	}

	private class ReferencePoint {
		public Point m_p = new Point();
		public ReferencePointListener m_listener = null;
		private Point m_dirty = new Point();
		
		public ReferencePoint( Point p, ReferencePointListener l ) {
			m_p = new Point(p);
			m_listener = l;
			clearDirty();
		}
		
		public boolean isDirty() {
			return !(m_p.equals(m_dirty));
		}
		
		public void clearDirty() {
			m_dirty.x = m_p.x;
			m_dirty.y = m_p.y;
		}
	}
}